/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.context;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import de.unibremen.informatik.tdki.grind.utilities.ELUtils;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectIntersectionOf;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;

/**
 *
 * @author seylan
 */
public class OmegaContext {

    private Context context;

    // This one maps from concepts to GCIs with concept in lhs:
    private Multimap<OWLClassExpression, OWLSubClassOfAxiom> conceptToGcis;
    // This one maps from a GCI to sets of concepts. From each set, at least one
    // concept must be present for the GCI to be able to fire.
    private Multimap<OWLSubClassOfAxiom, Set<OWLClassExpression>> gciToConcepts;

    private Set<OWLObjectIntersectionOf> conjunctionInLhsExistentials;

    private Map<Set<OWLClassExpression>, Set<OWLClassExpression>> cachedConsequences;

    public OmegaContext(Context c) {
        this.context = c;
        collectExistentialsWithComplexFillers();
        createConceptToGCIMappings();
        cachedConsequences = new HashMap<>(10000);
    }

    public Context getContext() {
        return context;
    }

    private void collectExistentialsWithComplexFillers() {
        conjunctionInLhsExistentials = new HashSet<>();
        for (OWLSubClassOfAxiom gci : context.getOriginalOnt().getGCIs()) {
            breakComplexLhs(gci.getSubClass());
        }
    }

    private void breakComplexLhs(OWLClassExpression c) {
        for (OWLClassExpression tlc : ELUtils.getTlcs(c)) {
            if (tlc instanceof OWLObjectSomeValuesFrom) {
                OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) tlc;
                if (!(ex.getFiller() instanceof OWLClass)) {
                    if (!(ex.getFiller() instanceof OWLObjectSomeValuesFrom)) { // conjunction
                        conjunctionInLhsExistentials.add((OWLObjectIntersectionOf) ex.getFiller());
                    }
                    breakComplexLhs(ex.getFiller());
                }
            }
        }
    }

    /**
     * Generate the two mappings from concepts to GCIs and vice versa, which is
     * useful for chasing (i.e., consequence calculations). Generate mapping
     * from all concepts C in sub(T) to GCIs C and D => E. Has been extended for
     * role hierarchies: If C is \exists r.F and s => r, then we need a mapping
     * from \exists s.F. So, establish the map for all s with \Tmc \models s =>
     * r.
     *
     * Generate mapping from GCI to sets of concepts. If we have at least one
     * concept from every set, the GCI can fire.
     */
    private void createConceptToGCIMappings() {
        conceptToGcis = HashMultimap.create();
        gciToConcepts = HashMultimap.create();

        // For all GCIs:
        for (OWLAxiom axiom : context.getOriginalOnt().getGCIs()) {
            OWLSubClassOfAxiom gci = (OWLSubClassOfAxiom) axiom;

            // Check all lhs TLCs:
            for (OWLClassExpression lhsTlc : ELUtils.getTlcs(gci.getSubClass())) {
                if (lhsTlc instanceof OWLObjectSomeValuesFrom) {

                    // For existential restrictions \exists r.F, we have to generate 
                    // \exists s.F for all s:
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) lhsTlc;
                    Set<OWLClassExpression> setOfConstructedNewExistentials = new HashSet<>();
                    for (OWLObjectPropertyExpression subrole : context.getReasoning().getEntailingRoles(ex.getProperty())) {
                        OWLClassExpression constructedNewExistential = context.getFactory().getOWLObjectSomeValuesFrom(subrole, ex.getFiller());
                        if (context.getOriginalOnt().getSubconcepts().contains(constructedNewExistential)) {
                            conceptToGcis.put(constructedNewExistential, gci);
                            setOfConstructedNewExistentials.add(constructedNewExistential);
                        }
                    }
                    gciToConcepts.put(gci, setOfConstructedNewExistentials);
                } else {
                    // We have a concept name, no role hierarchy stuff needed:
                    conceptToGcis.put(lhsTlc, gci);
                    Set<OWLClassExpression> setOfOneConceptName = new HashSet<>();
                    setOfOneConceptName.add(lhsTlc);
                    gciToConcepts.put(gci, setOfOneConceptName);
                }
            }
        }
//        System.out.println("conceptToGcis: " + conceptToGcis.toString()
//                .replace("http://www.semanticweb.org/owlapi/ontologies/ontology#", "")
//                .replace("], ", "],\n"));
//        System.out.println("gciToConcepts: " + gciToConcepts.toString()
//                .replace("http://www.semanticweb.org/owlapi/ontologies/ontology#", "")
//                .replace("], ", "],\n"));
//        System.exit(1);
    }

    @Deprecated
    private void createConceptToGCIMappingsOnlyLhsTlcs() {
        conceptToGcis = HashMultimap.create();
        gciToConcepts = HashMultimap.create();

        // For all GCIs:
        for (OWLAxiom axiom : context.getOriginalOnt().getGCIs()) {
            OWLSubClassOfAxiom gci = (OWLSubClassOfAxiom) axiom;

            // Check all lhs TLCs:
            for (OWLClassExpression lhsTlc : ELUtils.getTlcs(gci.getSubClass())) {
                if (lhsTlc instanceof OWLObjectSomeValuesFrom) {

                    // For existential restrictions \exists r.F, we have to generate 
                    // \exists s.F for all s:
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) lhsTlc;
                    Set<OWLClassExpression> setOfConstructedNewExistentials = new HashSet<>();
                    for (OWLObjectPropertyExpression subrole : context.getReasoning().getEntailingRoles(ex.getProperty())) {
                        OWLClassExpression constructedNewExistential = context.getFactory().getOWLObjectSomeValuesFrom(subrole, ex.getFiller());
                        if (context.getOriginalOnt().getSubconcepts().contains(constructedNewExistential)) {
                            conceptToGcis.put(constructedNewExistential, gci);
                            setOfConstructedNewExistentials.add(constructedNewExistential);
                        }
                    }
                    gciToConcepts.put(gci, setOfConstructedNewExistentials);
                } else {
                    // We have a concept name, no role hierarchy stuff needed:
                    conceptToGcis.put(lhsTlc, gci);
                    Set<OWLClassExpression> setOfOneConceptName = new HashSet<>();
                    setOfOneConceptName.add(lhsTlc);
                    gciToConcepts.put(gci, setOfOneConceptName);
                }
            }
        }
    }

    /**
     * Close a set of concepts S under consequence wrt the ontology.
     *
     * @param begin
     * @return
     */
    public Set<OWLClassExpression> closeUnderConsequence(Set<OWLClassExpression> begin) {
        if (cachedConsequences.containsKey(begin)) {
            return cachedConsequences.get(begin);
        }
        int previousSize = 0;
        Set<OWLClassExpression> result = new HashSet<>(begin);
        if (context.getOriginalOnt().getSubconcepts().contains(context.topConcept())) {
            result.add(context.topConcept());
        }

        Set<OWLSubClassOfAxiom> applicableGcis = new HashSet<>();

        // We iterate as long as we can extend result, i.e. as long as we need
        // to saturate the set of consequences.
        while (result.size() != previousSize) {
            previousSize = result.size();

            // Apply gcis from the TBox if possible:
            applyGcisFromOntology(result, applicableGcis);

            // Add the implied conjunctions to con, i.e.,
            // for A, B in S: add A \and B to con if (A \and B) is found in 
            // a complex lhs.
            addImpliedConjunctions(result);

            // Add implied existential restrictions to con, i.e., for \exists r.A in S,
            // and ELK decided T \models A => B: add \exists r.B to con.
            addImpliedExistentials(result);
        }

        cachedConsequences.put(begin, result);
        return result;
    }

    /**
     * Apply GCIs from the TBox if possible.
     */
    private void applyGcisFromOntology(Set<OWLClassExpression> result, Set<OWLSubClassOfAxiom> applicableGcis) {

        // Take the union of all relevant gcis:
        for (OWLClassExpression c : result) {
            applicableGcis.addAll(conceptToGcis.get(c));
        }

        // TODO: Think about how to prevent the same GCIs from being applied
        //       for the same reasons repeatedly.
        // Check each of them if all prerequisites are there to fire:
        for (OWLSubClassOfAxiom gci : applicableGcis) {

            // Check for every set in this GCI's preliminaries if we
            // have an element of the set in result:
            boolean gciCanFire = true;
            for (Set<OWLClassExpression> setOfConceptsOneNeeded : gciToConcepts.get(gci)) {
                boolean thisConjunctHasWitness = false;
                for (OWLClassExpression conceptInCon : result) {
                    if (setOfConceptsOneNeeded.contains(conceptInCon)) {
                        thisConjunctHasWitness = true;
                        break;
                    }
                }
                if (!thisConjunctHasWitness) {
                    gciCanFire = false;
                    break;
                }
            }
            if (gciCanFire) {
                result.addAll(ELUtils.getTlcs(gci.getSuperClass()));
            }
        }
    }

    /**
     * Add the implied conjunctions to con, i.e., for A, B in S: add A \and B to
     * con if (A \and B) is found in a complex lhs. This actually runs over all
     * deep lhs conjunctions and looks whether its conjuncts are present in con.
     */
    private void addImpliedConjunctions(Set<OWLClassExpression> result) {
        for (OWLObjectIntersectionOf conjunction : conjunctionInLhsExistentials) {
            if (result.containsAll(ELUtils.getTlcs(conjunction))) {
                result.add(conjunction);
            }
        }
    }

    /**
     * Add implied existential restrictions to con, i.e., for \exists r.C in S,
     * and ELK decided T \models C => D: add \exists r.D to con.
     */
    private void addImpliedExistentials(Set<OWLClassExpression> result) {
        Set<OWLClassExpression> impliedExistentials = new HashSet<>();

        // For everything in con so far:
        for (OWLClassExpression c : result) {
            // If it's an existential \exists r.C...
            if (c instanceof OWLObjectSomeValuesFrom) {
                OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) c;

                // Get every concept D that is implied by C \wrt T (using ELK):
                for (OWLClassExpression d : context.getReasoning().getEntailedConcepts(ex.getFiller())) {
                    // For sanity's sake: Don't generate new existential with same filler, please:
                    if (d.equals(ex.getFiller())) {
                        continue;
                    }

                    // Get all superroles of r, i.e. roles s with T \models r => s:
                    for (OWLObjectPropertyExpression superrole : context.getReasoning().getEntailedRoles(ex.getProperty())) {
                        // Construct \exists s.D. If it's in sub(T), add to con:
                        OWLClassExpression complexExistential = context.getFactory().getOWLObjectSomeValuesFrom(superrole, d);
                        if (context.getOriginalOnt().getSubconcepts().contains(complexExistential)) {
                            impliedExistentials.add(complexExistential);
                        }
                    }
                }
            }
        }
        result.addAll(impliedExistentials);
    }

    /**
     * This is the original implementation. Add implied existential restrictions
     * to con, i.e., for \exists r.A in S, and ELK decided T \models A => C: add
     * \exists r.C to con.
     */
    private void addImpliedExistentials2(Set<OWLClassExpression> result) {
        Set<OWLClassExpression> impliedExistentials = new HashSet<>();

        // For everything in con so far:
        for (OWLClassExpression c : result) {
            // If it's an existential...
            if (c instanceof OWLObjectSomeValuesFrom) {
                OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) c;
                // And a non-complex one, i.e. \exists r.A:
                if (ex.getFiller() instanceof OWLClass) {
                    // Get every concept C that is implied by A \wrt T (using ELK):
                    for (OWLClassExpression d : context.getReasoning().getEntailedConcepts(ex.getFiller())) {
                        // For sanity's sake: Don't generate new existential with same filler, please:
                        if (d.equals(ex.getFiller())) {
                            continue;
                        }
                        // Construct \exists r.C. If it's in sub(T), add to con:
                        OWLClassExpression complexExistential = context.getFactory().getOWLObjectSomeValuesFrom(ex.getProperty(), d);
                        if (context.getOriginalOnt().getSubconcepts().contains(complexExistential)) {
                            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                            System.err.println("IT FINALLY HAPPENED. EXISTENTIAL ADDED!");
                            impliedExistentials.add(complexExistential);
                        }
                    }
                }
            }
        }
        result.addAll(impliedExistentials);
    }

}
