/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.context;

import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.utilities.ELUtils;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.ClassBasedEdgeFactory;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;

/**
 *
 * @author seylan
 */
public class OriginalOntology {

    private final OWLOntologyManager manager;
    private final Set<OWLSubClassOfAxiom> gcis;
    private final Set<OWLSubObjectPropertyOfAxiom> roleInclusions;
    private final Map<OWLClassExpression, Integer> subconcepts2Int;
    private final Comparator<OWLClassExpression> comparator;
    private final Set<OWLClassExpression> subconcepts;

    public OriginalOntology(OWLOntology ont, OWLOntologyManager manager) {
        this.manager = manager;
//        if (checkForAcyclicity(ont)) {
//            System.out.println("Ontology is acyclic. No further processing needed.");
//            System.exit(0);
//        }

        //collectInfoAboutOntology(ont);
        
        eliminateDomainRestrictions(ont);
        eliminateRangeRestrictions(ont);
        
        convertEquivalentClassToInclusions(ont);
        convertEquivalentRoleToInclusions(ont);
        gcis = new HashSet<>(ont.getAxioms(AxiomType.SUBCLASS_OF));
        roleInclusions = new HashSet<>(ont.getAxioms(AxiomType.SUB_OBJECT_PROPERTY));

        this.subconcepts = ELUtils.getSubconcepts(gcis);
        this.subconcepts2Int = new HashMap<>();
        int i = 0;
        for (OWLClassExpression concept : subconcepts) {
            subconcepts2Int.put(concept, i++);
        }
        comparator = new ConceptComparator(subconcepts2Int);
    }

    public Set<OWLSubClassOfAxiom> getGCIs() {
        return gcis;
    }

    public Set<OWLSubObjectPropertyOfAxiom> getRoleInclusions() {
        return roleInclusions;
    }

    public Set<OWLClassExpression> getSubconcepts() {
        return subconcepts;
    }

    public Comparator<OWLClassExpression> getComparator() {
        return comparator;
    }

    private void eliminateDomainRestrictions(OWLOntology ontology) {
        final OWLDataFactory factory = OWLManager.getOWLDataFactory();
        for (OWLObjectPropertyDomainAxiom domainRestr : ontology.getAxioms(AxiomType.OBJECT_PROPERTY_DOMAIN)) {
            OWLClassExpression concept = domainRestr.getDomain();
            OWLObjectPropertyExpression role = domainRestr.getProperty();
            
            manager.removeAxiom(ontology, domainRestr);
            OWLObjectSomeValuesFrom exRTop = factory.getOWLObjectSomeValuesFrom(role, factory.getOWLThing());
            OWLSubClassOfAxiom exRTopImpliesC = factory.getOWLSubClassOfAxiom(exRTop, concept);
            manager.addAxiom(ontology, exRTopImpliesC);
            
//            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//            System.err.println("Domain restriction:\n" + StringTools.shorten(domainRestr));
//            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//            System.err.println("exRTopImpliesC:\n" + StringTools.shorten(exRTopImpliesC));
        }
    }
    private void eliminateRangeRestrictions(OWLOntology ontology) {
//        final OWLDataFactory factory = OWLManager.getOWLDataFactory();
//        for (OWLObjectPropertyRangeAxiom domainRestr : ontology.getAxioms(AxiomType.OBJECT_PROPERTY_RANGE)) {
//            OWLClassExpression concept = domainRestr.getDomain();
//            OWLObjectPropertyExpression role = domainRestr.getProperty();
//            
//            manager.removeAxiom(ontology, domainRestr);
//            OWLObjectSomeValuesFrom exRTop = factory.getOWLObjectSomeValuesFrom(role, factory.getOWLThing());
//            OWLSubClassOfAxiom exRTopImpliesC = factory.getOWLSubClassOfAxiom(exRTop, concept);
//            manager.addAxiom(ontology, exRTopImpliesC);
            
//            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//            System.err.println("Domain restriction:\n" + StringTools.shorten(domainRestr));
//            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//            System.err.println("exRTopImpliesC:\n" + StringTools.shorten(exRTopImpliesC));
//        }
    }
    
    private void convertEquivalentClassToInclusions(OWLOntology ontology) {
        Set<OWLAxiom> axiomsToRemove = new HashSet<>();
        for (OWLAxiom axiom : ontology.getAxioms(AxiomType.EQUIVALENT_CLASSES)) {
            OWLEquivalentClassesAxiom eqAx = (OWLEquivalentClassesAxiom) axiom;
            manager.addAxioms(ontology, eqAx.asOWLSubClassOfAxioms());
            axiomsToRemove.add(eqAx);
        }
        manager.removeAxioms(ontology, axiomsToRemove);
    }

    private void convertEquivalentRoleToInclusions(OWLOntology ontology) {
        Set<OWLAxiom> axiomsToRemove = new HashSet<>();
        for (OWLAxiom axiom : ontology.getAxioms(AxiomType.EQUIVALENT_OBJECT_PROPERTIES)) {
            OWLEquivalentObjectPropertiesAxiom eqAx = (OWLEquivalentObjectPropertiesAxiom) axiom;
            manager.addAxioms(ontology, eqAx.asSubObjectPropertyOfAxioms());
            axiomsToRemove.add(eqAx);
        }
        manager.removeAxioms(ontology, axiomsToRemove);
    }

    public Set<OWLClass> getConceptNames() {
        return ELUtils.getConceptNames(gcis);
    }

    public Set<OWLObjectProperty> getRoleNames() {
        return ELUtils.getRoleNames(gcis);
    }

    public static boolean collectInfoAboutOntology(OWLOntology ont) {

        int numberOfSubClassOfAxioms = 0;
        int numberOfEquivalenceAxioms = 0;
        //int CNsUsedInEqEqOrEqSC = 0;

        int numberOfABSCAxioms = 0;
        int numberOfCASCAxioms = 0;
        int numberOfACSCAxioms = 0;
        int numberOfCDSCAxioms = 0;

        int numberOfABEqAxioms = 0;
        int numberOfACEqAxioms = 0;
        int numberOfCAEqAxioms = 0;
        int numberOfCDEqAxioms = 0;

        Set<OWLClass> definedConceptnames = new HashSet<>();

        DirectedGraph cyclicityGraph = new DefaultDirectedGraph(
                new ClassBasedEdgeFactory<GammaTuple, DefaultEdge>(DefaultEdge.class));

        boolean uniqueNames = true;
        boolean complexLhs = false;

        for (OWLSubClassOfAxiom axiom : ont.getAxioms(AxiomType.SUBCLASS_OF)) {
            numberOfSubClassOfAxioms++;
            OWLClassExpression lhs = axiom.getSubClass();
            OWLClassExpression rhs = axiom.getSuperClass();

            if ((lhs instanceof OWLClass) && (rhs instanceof OWLClass)) {
                numberOfABSCAxioms++;
                if (!definedConceptnames.add((OWLClass) lhs)) {
                    uniqueNames = false;
                }
            } else if ((lhs instanceof OWLClass) && !(rhs instanceof OWLClass)) {
                numberOfACSCAxioms++;
                if (!definedConceptnames.add((OWLClass) lhs)) {
                    uniqueNames = false;
                }
            } else if ((rhs instanceof OWLClass) && !(lhs instanceof OWLClass)) {
                complexLhs = true;
                numberOfCASCAxioms++;
            } else {
                complexLhs = true;
                numberOfCDSCAxioms++;
            }
            if (lhs instanceof OWLClass) {
                cyclicityGraph.addVertex((OWLClass) lhs);
                for (OWLClass rhsCN : ELUtils.getCNsFromConcept(rhs)) {
                    cyclicityGraph.addVertex(rhsCN);
                    cyclicityGraph.addEdge((OWLClass) lhs, rhsCN);
                }
            }
        }

        for (OWLEquivalentClassesAxiom axiom : ont.getAxioms(AxiomType.EQUIVALENT_CLASSES)) {

            numberOfEquivalenceAxioms++;
            OWLSubClassOfAxiom decomposedAxiom = axiom.asOWLSubClassOfAxioms().iterator().next();
            OWLClassExpression lhs = decomposedAxiom.getSubClass();
            OWLClassExpression rhs = decomposedAxiom.getSuperClass();
            if ((lhs instanceof OWLClass) && (rhs instanceof OWLClass)) {
                numberOfABEqAxioms++;
            } else if ((lhs instanceof OWLClass) && !(rhs instanceof OWLClass)) {
                numberOfACEqAxioms++;
                if (!definedConceptnames.add((OWLClass) lhs)) {
                    uniqueNames = false;
                }
            } else if ((rhs instanceof OWLClass) && !(lhs instanceof OWLClass)) {
                numberOfCAEqAxioms++;
                if (!definedConceptnames.add((OWLClass) rhs)) {
                    uniqueNames = false;
                }
                // flip C=A into A=C:
                cyclicityGraph.addVertex((OWLClass) rhs);
                for (OWLClass lhsCN : ELUtils.getCNsFromConcept(lhs)) {
                    cyclicityGraph.addVertex(lhsCN);
                    cyclicityGraph.addEdge((OWLClass) rhs, lhsCN);
                }
            } else {
                numberOfCDEqAxioms++;
                complexLhs = true;
            }
            if (lhs instanceof OWLClass) {
                cyclicityGraph.addVertex((OWLClass) lhs);
                for (OWLClass rhsCN : ELUtils.getCNsFromConcept(rhs)) {
                    cyclicityGraph.addVertex(rhsCN);
                    cyclicityGraph.addEdge((OWLClass) lhs, rhsCN);
                }
            }
        }
        System.out.println("  Number of SubClassOf axioms:          " + numberOfSubClassOfAxioms);
        System.out.println("    Number of A => B axioms:            " + numberOfABSCAxioms);
        System.out.println("    Number of A => C axioms:            " + numberOfACSCAxioms);
        System.out.println("    Number of C => A axioms:            " + numberOfCASCAxioms);
        System.out.println("    Number of C => D axioms:            " + numberOfCDSCAxioms);
        System.out.println("  Number of Equivalence axioms:         " + numberOfEquivalenceAxioms);
        System.out.println("    Number of A = B axioms:             " + numberOfABEqAxioms);
        System.out.println("    Number of A = C axioms:             " + numberOfACEqAxioms);
        System.out.println("    Number of C = A axioms:             " + numberOfCAEqAxioms);
        System.out.println("    Number of C = D axioms:             " + numberOfCDEqAxioms);
        System.out.println("  Number of defined CNs:                " + definedConceptnames.size());
        System.out.println("  Unique names:                         " + uniqueNames);
        System.out.println("  Complex lhs:                          " + complexLhs);

        CycleDetector cd = new CycleDetector(cyclicityGraph);
        if (uniqueNames && !complexLhs && !cd.detectCycles()) {
            System.out.println("  Syntactical analysis: Acyclic!!");
            return true;
        } else {
            System.out.println("  Syntactical analysis: Cyclic.");
            return false;
        }
    }
}
