/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.context;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import de.unibremen.informatik.tdki.grind.utilities.ELUtils;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jgrapht.alg.TransitiveClosure;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleDirectedGraph;
import org.semanticweb.elk.owlapi.ElkReasonerFactory;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.ClassExpressionType;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataFactory;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;
import org.semanticweb.owlapi.model.OWLSubObjectPropertyOfAxiom;
import org.semanticweb.owlapi.reasoner.Node;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;

/**
 *
 * @author seylan
 */
public class Reasoning {

    private int conceptNameCounter;
    public final static String URI = "http://www.informatik.uni-bremen.de/tdki#ConceptName";
    private final OWLDataFactory factory;
    private Set<OWLSubClassOfAxiom> newGcis;
    private Set<OWLSubObjectPropertyOfAxiom> roleInclusions;
    private BiMap<OWLClassExpression, OWLClassExpression> lhsExistentialFillerToCNBiMap, rhsExistentialFillerToCNBiMap;
    private final Context context;

    /**
     * The following stores C1 -> C2 for T \models C1 \sqsubseteq C2, where Ci
     * is
     *
     * 1. a concept name from the TBox
     *
     * 2. the filler of an existential restriction that is a tlc of the rhs of
     * an axiom in T
     *
     * 3. the filler of an existential restriction that is a subconcept of a lhs
     * of an axiom in T
     */
    private HashMultimap<OWLClassExpression, OWLClassExpression> entailedConcepts;
    /**
     * The following is basically the inverse mapping of the one above
     */
    private HashMultimap<OWLClassExpression, OWLClassExpression> entailingConcepts;

    /**
     * Stores r -> s for T \models r \sqsubseteq s.
     */
    private HashMultimap<OWLObjectPropertyExpression, OWLObjectPropertyExpression> entailedRoles;
    /**
     * Stores s -> r for T \models r \sqsubseteq s.
     */
    private HashMultimap<OWLObjectPropertyExpression, OWLObjectPropertyExpression> entailingRoles;

    private Multimap<OWLClassExpression, OWLSubClassOfAxiom> relevantAxioms;

    protected Reasoning(Context context, Set<OWLSubClassOfAxiom> originalOntology, Set<OWLSubObjectPropertyOfAxiom> roleInclusions) {
        this.context = context;
        factory = OWLManager.getOWLDataFactory();
        normalize(originalOntology, roleInclusions);
        createEntailingMaps(roleInclusions);
        computeRelevantAxioms(originalOntology);
    }

    private void normalize(Set<OWLSubClassOfAxiom> originalGcis, Set<OWLSubObjectPropertyOfAxiom> roleInclusions) {
        conceptNameCounter = 0;
        newGcis = new HashSet<>();
        this.roleInclusions = roleInclusions;
        lhsExistentialFillerToCNBiMap = HashBiMap.create();
        rhsExistentialFillerToCNBiMap = HashBiMap.create();
        for (OWLSubClassOfAxiom gci : originalGcis) {
            // Go over the rhs of the GCI:
            for (OWLClassExpression tlc : ELUtils.getTlcs(gci.getSuperClass())) {
                if (tlc instanceof OWLObjectSomeValuesFrom) {
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) tlc;
                    if (!(ex.getFiller() instanceof OWLClass) && !rhsExistentialFillerToCNBiMap.containsKey(ex.getFiller())) {
                        OWLClassExpression newConceptName = factory.getOWLClass(IRI.create(URI + conceptNameCounter++));
                        rhsExistentialFillerToCNBiMap.put(ex.getFiller(), newConceptName);
                        OWLSubClassOfAxiom newGCI = factory.getOWLSubClassOfAxiom(newConceptName, ex.getFiller());
                        newGcis.add(newGCI);
                    }
                }
            }
            // Go over the lhs of the GCI:
            for (OWLClassExpression tlc : ELUtils.getSubconcepts(gci.getSubClass())) {
                if (tlc instanceof OWLObjectSomeValuesFrom) {
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) tlc;
                    if (!(ex.getFiller() instanceof OWLClass) && !lhsExistentialFillerToCNBiMap.containsKey(ex.getFiller())) {
                        OWLClassExpression newConceptName = factory.getOWLClass(IRI.create(URI + conceptNameCounter++));
                        lhsExistentialFillerToCNBiMap.put(ex.getFiller(), newConceptName);
                        OWLSubClassOfAxiom ax = factory.getOWLSubClassOfAxiom(ex.getFiller(), newConceptName);
                        newGcis.add(ax);
                    }
                }
            }
        }
        newGcis.addAll(originalGcis);
    }

    public boolean hasRelevantAxiom(OWLClassExpression c) {
        return relevantAxioms.containsKey(c);
    }

    public Collection<OWLSubClassOfAxiom> getRelevantAxioms(OWLClassExpression concept) {
        return relevantAxioms.get(concept);
    }

    /**
     * For all GCIs, get rhs TLCs C and map them to their GCI. This will serve
     * later to easily get all GCIs that may be applied backwards when
     * encountering the concept C in a rewriting.
     *
     * For role hierarchies, when we find a tlc \exists r.F in a rhs, we add
     * mappings for \exists s.G for all entailed concepts F => G and entailed
     * roles r => s using our prior reasoning.
     */
    private void computeRelevantAxioms(Set<OWLSubClassOfAxiom> originalOntology) {
        relevantAxioms = HashMultimap.create();
        for (OWLSubClassOfAxiom gci : originalOntology) {
            for (OWLClassExpression c : ELUtils.getTlcs(gci.getSuperClass())) {

                if (c instanceof OWLClass) {
                    relevantAxioms.put(c, gci);

                } else {
                    OWLObjectSomeValuesFrom existential = (OWLObjectSomeValuesFrom) c;

                    // Build up all existentials that are entailed, using
                    // entailedConcepts and entailedRoles.
                    for (OWLClassExpression g : getEntailedConcepts(existential.getFiller())) {
                        for (OWLObjectPropertyExpression role : getEntailedRoles(existential.getProperty())) {
                            OWLObjectSomeValuesFrom newExistential = factory.getOWLObjectSomeValuesFrom(role, g);
                            if (context.getOriginalOnt().getSubconcepts().contains(newExistential)) {
                                relevantAxioms.put(newExistential, gci);
                            }
                        }
                    }
                }
            }
        }
    }

    private void createEntailingMaps(Set<OWLSubObjectPropertyOfAxiom> roleInclusions) {
        OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
        OWLOntology ontology;
        try {
            // Ugly typecast below but we can not avoid it: 
            // manager.createOntology() needs a Set of OWLAxioms.
            Set<OWLAxiom> copyOfAxioms = new HashSet<>((Set<OWLAxiom>) (Set<? extends OWLAxiom>) newGcis);
            copyOfAxioms.addAll((Set<? extends OWLAxiom>) this.roleInclusions);
            ontology = manager.createOntology(copyOfAxioms);
        } catch (OWLOntologyCreationException ex) {
            throw new RuntimeException(ex);
        }
        Logger.getLogger("org.semanticweb.elk").setLevel(Level.ERROR);
        OWLReasonerFactory reasonerFactory = new ElkReasonerFactory();
        OWLReasoner reasoner = reasonerFactory.createReasoner(ontology);

        entailingConcepts = HashMultimap.create();
        entailedConcepts = HashMultimap.create();
        entailingRoles = HashMultimap.create();
        entailedRoles = HashMultimap.create();

        // Do concept  entailment calculations.
        for (OWLClassExpression c : ELUtils.getSubconcepts(newGcis)) {

            // Remark: we have to add all occurring roles in T, they may not appear in RIs.
            if (c.getClassExpressionType() == ClassExpressionType.OBJECT_SOME_VALUES_FROM) {
                OWLObjectPropertyExpression role = ((OWLObjectSomeValuesFrom) c).getProperty();
                entailedRoles.put(role, role);
                entailingRoles.put(role, role);
            }

            for (Node<OWLClass> node : reasoner.getSuperClasses(c, false)) {
                for (OWLClass d : node.getEntities()) {
                    // c and d may be the newly introduced concept names
                    // we need to convert them to concepts from the original ontology
                    OWLClassExpression entailer = getOriginalConcept(c);
                    OWLClassExpression entailee = getOriginalConcept(d);
                    entailingConcepts.put(entailee, entailer);
                    entailedConcepts.put(entailer, entailee);
                }
            }
            entailingConcepts.put(getOriginalConcept(c), getOriginalConcept(c));
            entailedConcepts.put(getOriginalConcept(c), getOriginalConcept(c));
        }

        // Do role entailment calculations.
        SimpleDirectedGraph<OWLObjectPropertyExpression, DefaultEdge> roleGraph
                = new SimpleDirectedGraph(DefaultEdge.class);
        for (OWLSubObjectPropertyOfAxiom ri : roleInclusions) {
            OWLObjectPropertyExpression lhs = ri.getSubProperty().asOWLObjectProperty();
            OWLObjectPropertyExpression rhs = ri.getSuperProperty().asOWLObjectProperty();
            roleGraph.addVertex(lhs);
            roleGraph.addVertex(rhs);
            roleGraph.addEdge(lhs, rhs);
        }

        // Do transitive closure in role graph.
        TransitiveClosure.INSTANCE.closeSimpleDirectedGraph(roleGraph);

        // Fill mappings.
        // Do reflexive closure only in mapping, not allowed in SimpleDirectedGraph.
        for (OWLObjectPropertyExpression role : roleGraph.vertexSet()) {
            for (DefaultEdge edge : roleGraph.incomingEdgesOf(role)) {
                entailingRoles.put(role, roleGraph.getEdgeSource(edge));
            }
            entailingRoles.put(role, role);  // refl. closure
            for (DefaultEdge edge : roleGraph.outgoingEdgesOf(role)) {
                entailedRoles.put(role, roleGraph.getEdgeTarget(edge));
            }
            entailedRoles.put(role, role);  // refl. closure
        }

        // Terminate the worker threads used by the reasoner.
        reasoner.dispose();
        manager.removeOntology(ontology);
    }

    private OWLClassExpression getOriginalConcept(OWLClassExpression c) {
        if (lhsExistentialFillerToCNBiMap.inverse().containsKey(c)) {
            return lhsExistentialFillerToCNBiMap.inverse().get(c);
        } else if (rhsExistentialFillerToCNBiMap.inverse().containsKey(c)) {
            return rhsExistentialFillerToCNBiMap.inverse().get(c);
        }
        return c;
    }

    public Set<OWLClassExpression> getEntailingConcepts(OWLClassExpression concept) {
        return entailingConcepts.get(concept);
    }

    public boolean isInEntailingConceptsMap(OWLClassExpression concept) {
        return entailingConcepts.containsKey(concept);
    }

    public Set<OWLClassExpression> getEntailedConcepts(OWLClassExpression concept) {
        return entailedConcepts.get(concept);
    }

    public boolean isInEntailedConceptsMap(OWLClassExpression concept) {
        return entailedConcepts.containsKey(concept);
    }

    public Set<OWLObjectPropertyExpression> getEntailingRoles(OWLObjectPropertyExpression role) {
        return entailingRoles.get(role);
    }

    public Set<OWLObjectPropertyExpression> getEntailedRoles(OWLObjectPropertyExpression role) {
        return entailedRoles.get(role);
    }
}
