/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.control;

import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.datalog.GammaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.datalog.OmegaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaGraphCreator;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaSaturator;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.omegaconstruction.Omega;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaLoopChecker;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaTuple;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogEntry;
import de.unibremen.informatik.tdki.grind.statistics.OverviewLogger;
import de.unibremen.informatik.tdki.grind.statistics.OverviewLogEntry;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogger;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.cycle.DirectedSimpleCycles;
import org.jgrapht.alg.cycle.SzwarcfiterLauerSimpleCycles;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;

/**
 *
 * @author hansen
 */
public class GammaAndOmegaCaller {

    private static final int S = 1000; // 1 s = 1000 ms
    private static final int MIN = 60_000;
    private static final int H = 60 * 60_000;

    private static final int TIMEOUT = 30 * S;
//    private static int GAMMA_TIMEOUT = 15 * S;
//    private static int OMEGA_TIMEOUT = 15 * S;

    private OWLOntologyManager manager;
    private File file;
    private final String ontSource;
    private OWLOntology ont;
    private int noOfCNs;
    private Context context;
    private long tStartOfThisConcept;
    private final long t0;
    private Set<OWLClass> conceptNames;
    private PrintWriter output;
    private int numberOfGammaCycles;
    private int numberOfGammaCycleFOno, numberOfGammaCycleFOyes;
    private int numberOfGammaTimeouts;
    private int numberOfOmegaTimeouts;
    private OverviewLogEntry overviewLogEntry;
    private final boolean verbose;
    private final boolean logging;
    private final boolean norewrite;

    public GammaAndOmegaCaller(String ontSource, boolean logging, boolean verbose, boolean norewrite) {
        // Start global time count:
        t0 = System.currentTimeMillis();

        this.ontSource = ontSource;
        this.logging = logging;
        this.verbose = verbose;
        this.norewrite = norewrite;
    }

    /**
     *
     * @param startConceptString
     * @return return value for shell: 0=FO-rewr, 1=Not FO-rewr.
     * @throws OWLOntologyCreationException
     * @throws FileNotFoundException
     * @throws Exception
     */
    public int runSingleConceptname(String startConceptString)
            throws OWLOntologyCreationException, FileNotFoundException, Exception {

        overviewLogEntry = new OverviewLogEntry(ontSource);
        openOntology(ontSource);

        OWLClass startConcept = null;
        boolean found = false;
        for (OWLClass concept : conceptNames) {
            if (StringTools.getFragment(concept.getIRI()).equals(startConceptString)) {
                startConcept = concept;
                found = true;
                break;
            }
        }
        if (!found) {
            System.err.println("grind: Argument " + startConceptString + " is not a concept name in " + ontSource + ". Aborting.");
            System.exit(-1);
        }

        tStartOfThisConcept = System.currentTimeMillis();
        DetailsLogEntry detailsLogEntry = new DetailsLogEntry();
        noOfCNs = 1;
        int retval = runGammaAndOmega(startConcept, 1, detailsLogEntry);
        if (verbose) {
            System.out.println("\nTotal time: " + (System.currentTimeMillis() - tStartOfThisConcept) + " ms.\n");
        }
        detailsLogEntry.setTimeSpent(System.currentTimeMillis() - tStartOfThisConcept);
        return retval;
    }

    public void runAllConceptnames()
            throws OWLOntologyCreationException, FileNotFoundException, Exception {

        overviewLogEntry = new OverviewLogEntry(ontSource);
        openOntology(ontSource);
        int currentClass = 0;

        noOfCNs = conceptNames.size();
        for (OWLClass concept : conceptNames) {
            tStartOfThisConcept = System.currentTimeMillis();
            currentClass++;
            System.out.println("\n" + StringTools.getFragment(concept.getIRI()) + ":\n");
            DetailsLogEntry detailsLogEntry = new DetailsLogEntry();

            runGammaAndOmega(concept, currentClass, detailsLogEntry);

            detailsLogEntry.setTimeSpent(System.currentTimeMillis() - tStartOfThisConcept);
            if (logging) {
                DetailsLogger.getInstance().writeLogEntry(detailsLogEntry);
            }
            System.out.println();
        }

        long timeNeeded = (System.currentTimeMillis() - t0);
        if (verbose) {
            System.out.println("\nTotal number of cyclic Gamma graphs: " + numberOfGammaCycles + ".");
            System.out.println("Total time: " + timeNeeded + " ms.");
        }
        if (logging) {
            overviewLogEntry.setNoOfGammaCycles(numberOfGammaCycles);
            overviewLogEntry.setNoOfGammaCycleFOno(numberOfGammaCycleFOno);
            overviewLogEntry.setNoOfGammaCycleFOyes(numberOfGammaCycleFOyes);
            overviewLogEntry.setNoOfGammaTimeouts(numberOfGammaTimeouts);
            overviewLogEntry.setNoOfOmegaTimeouts(numberOfOmegaTimeouts);
            overviewLogEntry.setTimeNeeded(timeNeeded);

            OverviewLogger.getInstance().writeLogEntry(overviewLogEntry);
        }
        if (verbose) {
            System.out.println(overviewLogEntry);
        }
    }

    private void openOntology(String ontSource) {

        if (verbose) {
            System.out.println("\n\nOpening ontology " + ontSource + "...");
        }
        manager = OWLManager.createOWLOntologyManager();
        file = new File(ontSource);
        try {
            ont = manager.loadOntologyFromOntologyDocument(file);
        } catch (Exception e) {
            System.err.println("Error: Could not load ontology: " + e);
            System.exit(-1);
        }
        context = new Context(ont, manager);
        manager.removeOntology(ont);

        if (logging) {
            overviewLogEntry.setNoOfAxioms(context.getOriginalOnt().getGCIs().size()
                    + context.getOriginalOnt().getRoleInclusions().size());
        }
        if (verbose) {
            System.out.println("  Normalising the ontology: " + (System.currentTimeMillis() - t0) + " ms.");
            System.out.println("\n  Size of sub(T):           " + context.getOriginalOnt().getSubconcepts().size());
        }
        conceptNames = context.getOriginalOnt().getConceptNames();
        noOfCNs = conceptNames.size();
        overviewLogEntry.setNoOfConceptnames(noOfCNs);
        overviewLogEntry.setNoOfRolenames(context.getOriginalOnt().getRoleNames().size());
    }

    private int runGammaAndOmega(OWLClass startConcept, int currentClass, DetailsLogEntry detailsLogEntry) throws Exception {

        detailsLogEntry.setConceptName(startConcept);

        if (verbose) {
            System.out.print("(" + currentClass + " / " + noOfCNs + ") ");
            System.out.println("Working on class " + startConcept.getIRI().getFragment());
        }

        // * Saturate Gamma:
        GammaSaturator gammaSat = new GammaSaturator(startConcept, detailsLogEntry, context);
        long usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
        gammaSat.saturate(TIMEOUT - usedUntilNow);

        // Main loop, do Gamma - Omega - Gamma - Omega...
        while (true) {

            // * if Gamma timeout:
            if (detailsLogEntry.isGammaTimeout()) {
                numberOfGammaTimeouts++;
                return -1;
            }

            // * if Gamma acyclic:
            Gamma gamma = gammaSat.getGamma();
            gamma.createGammaGraph();
            
            if (!gamma.isCyclic()) {
                if (!norewrite) {
                    GammaToDatalogRewriter dr = new GammaToDatalogRewriter(gamma, startConcept, context);
                    System.out.println(dr.gammaToDatalog());
                } else {
                    System.out.println("Concept " + startConcept.getIRI().getFragment() + " is FO-rewritable (Gamma acyclic).");
                }
                if (logging) {
                    detailsLogEntry.setCyclicGamma(false);
                    detailsLogEntry.setIsFOrewritable(true);
                }
                return 0;
            }

            // Increase number of Gamma cycles only once:
            if (!detailsLogEntry.isCyclicGamma()) {
                detailsLogEntry.setCyclicGamma(true);
                numberOfGammaCycles++;
            }

            detailsLogEntry.setNoOfGammaTuplesOnCycle(gamma.numberOfTuplesOnCycle());

            if (verbose) {
                System.out.println("Gamma graph:");
                System.out.println("  Number of vertices:     " + gamma.getGammaGraph().vertexSet().size());
                System.out.println("  Number of edges:        " + gamma.getGammaGraph().edgeSet().size());
                //System.out.print("Distr. of ex. restr. in S: ");
                //System.out.println("    " + countExistentials(gamma.getGammaTuples()));
                System.out.println("  # of vert. lying on cycle: " + gamma.numberOfTuplesOnCycle());
                System.out.println("  distr. of their ex. restr.:");
                System.out.println("    " + countExistentials(gamma.getTuplesOnCycle()));
                //countSimpleCycles(graph);  // VERY costly!!
                System.out.println();
            }

            // * Saturate Omega:
            OmegaSaturator omegaSat = new OmegaSaturator(gamma, detailsLogEntry, startConcept, norewrite, context);
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            try {
                omegaSat.saturate(TIMEOUT - usedUntilNow);
            } catch (OutOfMemoryError e) {
                System.err.println(e);
                omegaSat.dumpInternals();
                System.exit(-1);
            }
            Omega omega = omegaSat.getOmega();
            
            // * if Omega timeout:
            if (detailsLogEntry.isOmegaTimeout()) {
                numberOfOmegaTimeouts++;
                return -2;
            }

            detailsLogEntry.setNoOfOmegaTuples(omega.getOmegaTuples().size());

            // * if Omega cyclic:
            DirectedGraph<OmegaTuple, DefaultEdge> omegaGraph = omegaSat.getOmegaGraph();
            OmegaLoopChecker omegaLoopChecker = new OmegaLoopChecker(startConcept, omegaGraph, omegaSat);
            // To construct the Datalog rewriting, we only need the component 
            // of the graph that is reachable from root tuples. Note that 
            // we actually do the check for Omega loops here and get the 
            // reachable component for free.
            boolean omegaLoopFree = omegaLoopChecker.retainReachableComponentOrFail();
            if (!omegaLoopFree) {
                System.out.println("Concept " + startConcept.getIRI().getFragment()
                        + " is not FO-rewritable.");
                if (logging) {
                    detailsLogEntry.setIsFOrewritable(false);
                }
                numberOfGammaCycleFOno++;
                return 1;
            }

            // * if Omega acyclic and Gamma not premature:
            if (omegaLoopFree && !gammaSat.isPrematureReturnBecauseOfCycle()) {
                if (!norewrite) {
                    OmegaToDatalogRewriter dr = new OmegaToDatalogRewriter(omegaSat, startConcept, context);
                    System.out.println(dr.omegaToDatalog());
                } else {
                    System.out.println("Concept " + startConcept.getIRI().getFragment() + " is FO-rewritable (Omega cancelling).");
                }
                if (logging) {
                    detailsLogEntry.setIsFOrewritable(true);
                }
                numberOfGammaCycleFOyes++;
                return 0;
            }

            // * Resume Gamma saturation:
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            gammaSat.resumeSaturation(TIMEOUT - usedUntilNow);

        } // while true
    }

    /**
     * This gets all Gamma tuples that lie on a cycle.
     *
     * @param setOfTuplesOnCycle
     * @return
     */
    public static Map<Integer, Integer> countExistentials(Set<GammaTuple> setOfTuplesOnCycle) {

        Map<Integer, Integer> stat = new HashMap<>();

        for (GammaTuple tuple : setOfTuplesOnCycle) {

//            System.err.print("[DEBUG GammaAndOmegaCaller] ");
//            System.err.println("Gamma tuple on cycle: " + tuple);
            int numberOfExistentials = 0;
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    numberOfExistentials++;
                }
            }
            if (!stat.containsKey(numberOfExistentials)) {
                stat.put(numberOfExistentials, 1);
            } else {
                stat.put(numberOfExistentials, stat.get(numberOfExistentials) + 1);
            }
        }
        return stat;
    }

    private static Map<Integer, Integer> detailedListOutput(List<List<GammaTuple>> listOfList) {

        Map<Integer, Integer> stat = new HashMap<>();

        for (List list : listOfList) {
            if (!stat.containsKey(list.size())) {
                stat.put(list.size(), 1);
            } else {
                stat.put(list.size(), stat.get(list.size()) + 1);
            }
        }
        return stat;
    }

    /**
     * Count simple cycles in the given graph. This uses jGraphT's
     * SzwarcfiterLauerSimpleCycles, which might be the best implementation for
     * the job, but it is nonetheless EXTREMELY COSTLY to perform this!
     *
     * @param graph
     */
    private void countSimpleCycles(DirectedGraph graph) {
        DirectedSimpleCycles<GammaTuple, GammaGraphCreator.RoleEdge> tsc
                = new SzwarcfiterLauerSimpleCycles<>(graph);
        List<List<GammaTuple>> listOfList = tsc.findSimpleCycles();
        System.out.println("  # of simple cycles:        " + listOfList.size());
        for (List list : listOfList) {
            System.out.println("  # of tuples:               " + list.size());
        }
        System.out.println("  distr. of cycle sizes:");
        System.out.println("    " + detailedListOutput(listOfList));
        System.out.println("\n");
    }
}
