/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.datalog;

import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import java.util.HashMap;
import java.util.Map;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class GammaToDatalogRewriter {

    Gamma gamma;
    OWLClassExpression startConcept;
    Context context;
    
    /**
     * Provides methods to produce a Datalog rewriting from an acyclic set of 
     * Gamma tuples.
     * @param gamma
     * @param context
     * @param startConcept
     */
    public GammaToDatalogRewriter(Gamma gamma, OWLClassExpression startConcept, Context context) {
        this.gamma = gamma;
        this.startConcept = startConcept;
        this.context = context;
    }

    /**
     * Perform rewriting from set of Gamma tuples into Datalog rules. In the
     * paper, we get a disjunctive Datalog rule from a Gamma tuple (C, S):
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * \bigvee_{\Tmc\models s\sqsubseteq r, s\in \Sigma} ( s(x,y_{r,D}) \wedge
     * P_D(y_{r,D}) )
     *
     * Now we introduce auxiliary IDB predicates P_{C, \exists r.D} and get the
     * following:
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * P_{C, \exists r.D}(x)
     *
     * P_{C, \exists r.D}(x) :- s(x, y_{r,D}) \wedge P_D(y_{r,D}) for all \Tmc
     * \models s \sqsubseteq r
     *
     * @return
     */
    public String gammaToDatalog() {
        String result = "";

        // Start by analysing the C parts of the Gamma tuples.
        // Build up a mapping   GoalConcept --> "Q",
        //                     OtherConcept --> "AUX$1",
        //                    OtherConcept2 --> "AUX$2".
        Map<OWLClassExpression, String> idbMap = new HashMap<>();
        int idbCounter = 1;

        for (GammaTuple g : gamma.getGammaTuples()) {
            OWLClassExpression con = g.getConcept();
            if (!idbMap.containsKey(con)) {
                if (con == startConcept) {
                    idbMap.put(con, "Q");
                } else {
                    idbMap.put(con, "GammaFi$" + idbCounter++);
                }
            }
        }

        // We will need another set of IDBs to represent existentials in
        // S sets.
        Map<OWLClassExpression, String> sExIdbMap = new HashMap<>();

        // Now process all Gamma tuples.
        for (GammaTuple g : gamma.getGammaTuples()) {

            String additionalIDBRules = "";

            // Let (C, S) be our Gamma tuple.
            // Now construct the C IDB part (lhs).
            result += idbMap.get(g.getConcept()) + "(?0)  <-  ";
            for (OWLClassExpression con : g.getSetOfConcepts()) {
                if (con instanceof OWLClass) {
                    result += StringTools.getFragment(con.asOWLClass().getIRI()) + "(?0), ";
                } else if (con instanceof OWLObjectSomeValuesFrom) {
                    // Let's see if we already have an IDB for this:
                    if (idbMap.containsKey(con)) {
                        result += idbMap.get(con) + "(?0), ";
                    } else {
                        // Generate unique S existential IDB predicate on-the-fly:
                        String sExIdbName = "GammaEx$" + idbCounter++;
                        idbMap.put(con, sExIdbName);
                        result += sExIdbName + "(?0), ";
                        // Add additional IDB rules for this IDB:
                        OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) con;
                        for (OWLObjectPropertyExpression subrole
                                : context.getReasoning().getEntailingRoles(ex.getProperty())) {
                            additionalIDBRules += sExIdbName + "(?0)  <-  "
                                    + StringTools.getFragment(subrole.asOWLObjectProperty().getIRI()) + "(?0, ?1), "
                                    + idbMap.get(((OWLObjectSomeValuesFrom) con).getFiller()) + "(?1)\n";
                        }
                    }
                }
            }
            // Delete final ", " from result:
            result = result.substring(0, result.length() - 2);
            result += "\n" + additionalIDBRules;
        }
        return StringTools.uniq(result);
    }

}
