/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.datalog;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaTuple;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class TuplesetToDatalogRewriterOLD {

    Context context;

    /**
     * Provides methods to produce a Datalog rewriting from a set of Gamma/Omega
     * tuples.
     *
     * @param context
     */
    public TuplesetToDatalogRewriterOLD(Context context) {
        this.context = context;
    }

    /**
     * Perform rewriting from set of Gamma tuples into Datalog rules. In the
     * paper, we get a disjunctive Datalog rule from a Gamma tuple (C, S):
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * \bigvee_{\Tmc\models s\sqsubseteq r, s\in \Sigma} ( s(x,y_{r,D}) \wedge
     * P_D(y_{r,D}) )
     *
     * Now we introduce auxiliary IDB predicates P_{C, \exists r.D} and get the
     * following:
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * P_{C, \exists r.D}(x)
     *
     * P_{C, \exists r.D}(x) :- s(x, y_{r,D}) \wedge P_D(y_{r,D}) for all \Tmc
     * \models s \sqsubseteq r
     *
     * @param gamma
     * @param startConcept
     * @return
     */
    public String gammaToDatalog(Gamma gamma, OWLClassExpression startConcept) {
        String result = "";

        // Start by analysing the C parts of the Gamma tuples.
        // Build up a mapping   GoalConcept --> "Q",
        //                     OtherConcept --> "AUX$1",
        //                    OtherConcept2 --> "AUX$2".
        Map<OWLClassExpression, String> idbMap = new HashMap<>();
        int idbCounter = 1;

        for (GammaTuple g : gamma.getGammaTuples()) {
            OWLClassExpression con = g.getConcept();
            if (!idbMap.containsKey(con)) {
                if (con == startConcept) {
                    idbMap.put(con, "Q");
                } else {
                    idbMap.put(con, "GammaFi$" + idbCounter++);
                }
            }
        }

        // We will need another set of IDBs to represent existentials in
        // S sets.
        Map<OWLClassExpression, String> sExIdbMap = new HashMap<>();

        // Now process all Gamma tuples.
        for (GammaTuple g : gamma.getGammaTuples()) {

            String additionalIDBRules = "";

            // Let (C, S) be our Gamma tuple.
            // Now construct the C IDB part (lhs).
            result += idbMap.get(g.getConcept()) + "(?0)  <-  ";
            for (OWLClassExpression con : g.getSetOfConcepts()) {
                if (con instanceof OWLClass) {
                    result += StringTools.getFragment(con.asOWLClass().getIRI()) + "(?0), ";
                } else if (con instanceof OWLObjectSomeValuesFrom) {
                    // Let's see if we already have an IDB for this:
                    if (idbMap.containsKey(con)) {
                        result += idbMap.get(con) + "(?0), ";
                    } else {
                        // Generate unique S existential IDB predicate on-the-fly:
                        String sExIdbName = "GammaEx$" + idbCounter++;
                        idbMap.put(con, sExIdbName);
                        result += sExIdbName + "(?0), ";
                        // Add additional IDB rules for this IDB:
                        OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) con;
                        for (OWLObjectPropertyExpression subrole
                                : context.getReasoning().getEntailingRoles(ex.getProperty())) {
                            additionalIDBRules += sExIdbName + "(?0)  <-  "
                                    + StringTools.getFragment(subrole.asOWLObjectProperty().getIRI()) + "(?0, ?1), "
                                    + idbMap.get(((OWLObjectSomeValuesFrom) con).getFiller()) + "(?1)\n";
                        }
                    }
                }
            }
            // Delete final ", " from result:
            result = result.substring(0, result.length() - 2);
            result += "\n" + additionalIDBRules;
        }
        return StringTools.uniq(result);
    }

    // We need the following mappings: 
    //                (C, S, con, E, xcon) --> (C, _, con, _, _)
    //                (C, S, con, E, xcon) --> (C, S, con, _, _)
    //                (C, S, con, E, xcon) <-- (C, S, con, _, _)
    //   (C, S, con, E, xcon), \exists r.C --> (C, {\exists r.C}, con, _, _)
    private Set<OmegaTuple> setOfCAndCon = new HashSet<>();
    private Set<OmegaTuple> setOfCSAndCon = new HashSet<>();
    private Multimap<OmegaTuple, OmegaTuple> inverseFromCSAndCon = HashMultimap.create();
    private Set<OmegaTuple> setOfCConAndEx = new HashSet<>();
    private DirectedGraph<OmegaTuple, DefaultEdge> omegaGraph;
    // This will be the graph holding the \rightsquigarrow relation:
    DirectedGraph<OmegaTuple, LabelledEdge> squigOmegaGraph
            = new DefaultDirectedGraph<>(LabelledEdge.class);
    OWLClassExpression startConcept;
    private Map<OmegaTuple, String> cAndConTupleToIdbname = new HashMap<>();
    private Map<OmegaTuple, String> cConAndExTupleToIdbname = new HashMap<>();

    private OmegaTuple getCAndConTuple(OmegaTuple tuple) {
        return new OmegaTuple(tuple.getConcept(), null, tuple.getCon(), null, null);
    }

    private OmegaTuple getCSAndConTuple(OmegaTuple tuple) {
        return new OmegaTuple(tuple.getConcept(), tuple.getSetOfConcepts(), tuple.getCon(), null, null);
    }

    private OmegaTuple getCConAndExTuple(OmegaTuple tuple, OWLObjectSomeValuesFrom ex) {
        return new OmegaTuple(tuple.getConcept(), new TreeSet(Arrays.asList(ex)), tuple.getCon(), null, null);
    }

    /**
     * Contruct a non-recursive monadic Datalog rewriting from an acyclic Omega
     * set.
     *
     * @param startConcept
     * @param omegaSat
     * @return
     */
    public String omegaToDatalog(OWLClassExpression startConcept, OmegaSaturator omegaSat) {
        
        String result = "";
        this.startConcept = startConcept;
        // This is the graph that holds the selected successor relation:
        this.omegaGraph = omegaSat.getOmegaGraph();

        // Calculate the essential mappings:
        buildUpMappings();

        // Calculate the squig graph:
        constructSquigOmegaGraph();

        // We can now produce the Datalog rules of the rewriting.
        for (OmegaTuple tuple : squigOmegaGraph.vertexSet()) {

            // Let       (C, S, con, E, xcon) be our Omega tuple.
            // Get       (C, _, con, _, _   ):
            OmegaTuple cAndConTuple = getCAndConTuple(tuple);

            // Now construct the C_t/con_t IDB part (lhs):
            result += cAndConTupleToIdbname.get(cAndConTuple) + "(?0)  <-  ";

            String additionalIDBRules = "";

            // Construct a rewriting going over the S set:
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {

                // For concept names, directly put a predicate:
                if (concept instanceof OWLClass) {
                    result += StringTools.getFragment(concept.asOWLClass().getIRI()) + "(?0), ";

                    // For existentials, we need more effort:
                } else if (concept instanceof OWLObjectSomeValuesFrom) {
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) concept;
                    String idbName = cConAndExTupleToIdbname.get(getCConAndExTuple(tuple, ex));
                    result += idbName + "(?0), ";

                    // Then, for all squig successors of the tuple along \exists s.D
                    // (where s is subrole of r!)
                    for (LabelledEdge outgoingSquigEdge : squigOmegaGraph.outgoingEdgesOf(tuple)) {
                        if (context.getReasoning().getEntailingRoles(tuple.getSuccessor().getProperty())
                                .contains(outgoingSquigEdge.getLabel().getProperty())) {

                            OWLObjectPropertyExpression squigRole = outgoingSquigEdge.getLabel().getProperty();
                            OmegaTuple squigSuccessor = squigOmegaGraph.getEdgeTarget(outgoingSquigEdge);
                            additionalIDBRules += idbName + "(?0)  <-  "
                                    + StringTools.getFragment(squigRole.asOWLObjectProperty().getIRI())
                                    + "(?0, ?1), "
                                    + cAndConTupleToIdbname.get(getCAndConTuple(squigSuccessor))
                                    + "(?1)\n";
                        }
                    }
                }
            }
            // Delete final ", " from result:
            result = result.substring(0, result.length() - 2);
            result += "\n" + additionalIDBRules;
        }

        // As an easy size reduction, we eliminate duplicate lines:
        return StringTools.uniq(result);
    }

    /**
     * From the Omega graph, construct the squigOmegaGraph.
     */
    private void constructSquigOmegaGraph() {

        // Fill squig graph with all Omega vertices:
        for (OmegaTuple vertex : omegaGraph.vertexSet()) {
            squigOmegaGraph.addVertex(vertex);
        }

        // Iterate over all edges (source -> target) in the Omega graph
        // (i.e., target is a chosen successor [\exists s.D] of source).
        // For every source node, there might be several "brothers" that we need 
        // to also point to the target node in the squigOmegaGraph.
        // "Brothers" means equal C, equal S, equal con, and E = \exists r.D
        // with T \models s => r.
        for (DefaultEdge edge : omegaGraph.edgeSet()) {
            OmegaTuple source = omegaGraph.getEdgeSource(edge);
            OmegaTuple target = omegaGraph.getEdgeTarget(edge);

            // chosen successor = \exists s.D
            OWLObjectSomeValuesFrom chosenSuccessor = source.getSuccessor();

            OmegaTuple sourceCSAndCon = getCSAndConTuple(source);

            // For all sameCSAndCon tuples of source...
            for (OmegaTuple sameCSAndConTuple : inverseFromCSAndCon.get(sourceCSAndCon)) {

                /*
                 // We don't have to check ALL existentials in S of the sameCSCon tuple;
                 // the criterion is that the successor (\exists r.D) of the sameCSCon
                 // tuple has same D and superrole r of s!!
                 if (sameCSAndConTuple.getSuccessor().getFiller() == target.getConcept()
                 && context.getReasoning().getEntailedRoles(chosenSuccessor.getProperty())
                 .contains(sameCSAndConTuple.getSuccessor().getProperty())) {

                 // Label the edge with \exists s.D
                 LabelledEdge newEdge = new LabelledEdge(sameCSAndConTuple, target, chosenSuccessor);
                 squigOmegaGraph.addEdge(sameCSAndConTuple, target, newEdge);
                 }
                 */
                // For every concept in S...
                for (OWLClassExpression concept : sameCSAndConTuple.getSetOfConcepts()) {

                    // Check whether it's an existential...
                    if (concept instanceof OWLObjectSomeValuesFrom) {
                        OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) concept;
                        // Add edge to target if ex is \exists r.D with r superrole of s.
                        if (ex.getFiller() == target.getConcept()
                                && context.getReasoning().getEntailedRoles(chosenSuccessor.getProperty())
                                .contains(ex.getProperty())) {

                            // Label the edge with \exists s.D
                            LabelledEdge newEdge = new LabelledEdge(sameCSAndConTuple, target, chosenSuccessor);
                            squigOmegaGraph.addEdge(sameCSAndConTuple, target, newEdge);
                        }
                    }
                }
            }
        }
    }

    /**
     * Build up the essential mappings.
     */
    private void buildUpMappings() {

        for (OmegaTuple tuple : omegaGraph.vertexSet()) {

            OmegaTuple cAndConTuple = getCAndConTuple(tuple);
            setOfCAndCon.add(cAndConTuple);
            OmegaTuple cSAndConTuple = getCSAndConTuple(tuple);
            setOfCSAndCon.add(cSAndConTuple);
            inverseFromCSAndCon.put(cSAndConTuple, tuple);
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) concept;
                    setOfCConAndEx.add(getCConAndExTuple(tuple, ex));
                }
            }
        }

        // Map CAndCon tuples to integers. These will be used naming auxiliary predicates.
        //           (C, _, con, _, _) --> int
        int id = 1;
        for (OmegaTuple cAndConTuple : setOfCAndCon) {
            if (cAndConTuple.getConcept() == startConcept
                    && cAndConTuple.getCon().contains(startConcept)) {
                cAndConTupleToIdbname.put(cAndConTuple, "Q");
            } else {
                cAndConTupleToIdbname.put(cAndConTuple, "OmegaFi$" + id++);
            }
        }

        // Map CConAndEx tuples to integers.
        for (OmegaTuple cConAndExTuple : setOfCConAndEx) {
            cConAndExTupleToIdbname.put(cConAndExTuple, "OmegaEx$" + id++);
        }
    }

    private static class LabelledEdge<V> extends DefaultEdge {

        private final V source;
        private final V target;
        private final OWLObjectSomeValuesFrom label;

        public LabelledEdge(V source, V target, OWLObjectSomeValuesFrom label) {
            this.source = source;
            this.target = target;
            this.label = label;
        }

        public OWLObjectSomeValuesFrom getLabel() {
            return label;
        }

        @Override
        public String toString() {
//            return "(" + source + " ~~(" + label + ")~~ " + target + ")";
            return "( ~~(" + label + ")~~ )";
        }
    }

}
