/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.gammaconstruction;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multiset;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.ClassBasedEdgeFactory;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.DirectedPseudograph;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class GammaGraphCreator {

    private DirectedPseudograph<GammaTuple, GammaGraphCreator.RoleEdge> gammaGraph;
    private Multimap<OWLClassExpression, GammaTuple> concept2GammaTupleMap;
    Gamma gamma;
    OWLClass startConcept;

    public GammaGraphCreator(Gamma gamma, OWLClass startConcept) {
        this.gamma = gamma;
        this.startConcept = startConcept;
//        gammaGraph = new DefaultDirectedGraph<>(new ClassBasedEdgeFactory<GammaTuple, GammaGraphCreator.RoleEdge>(GammaGraphCreator.RoleEdge.class));
        gammaGraph = new DirectedPseudograph<>(new ClassBasedEdgeFactory<GammaTuple, GammaGraphCreator.RoleEdge>(GammaGraphCreator.RoleEdge.class));
        concept2GammaTupleMap = HashMultimap.create();
        createGraph();
    }

    /**
     * Creates a gammaGraph representation of the Gamma tuples if called.
     */
    private void createGraph() {

//        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//        System.err.println("Creating Gamma graph...");
        
        for (GammaTuple tuple : gamma.getGammaTuples()) {
            gammaGraph.addVertex(tuple);
            concept2GammaTupleMap.put(tuple.getConcept(), tuple);
        }

        // Iterate over tuples and seek role edges.
        // A Tuple (C,S) containing \exists r.D in S has edges to all tuples (D,_).
        for (GammaTuple tuple : gamma.getGammaTuples()) {
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    for (GammaTuple otherTuple : concept2GammaTupleMap.
                            get(((OWLObjectSomeValuesFrom) concept).getFiller())) {
                        gammaGraph.addEdge(tuple, otherTuple, new GammaGraphCreator.RoleEdge<>(tuple, otherTuple,
                                ((OWLObjectSomeValuesFrom) concept).getProperty()));
                    }
                }
            }
        }
    }

    public DirectedPseudograph getGraph() {
        return gammaGraph;
    }

    /**
     * Idea: We can only have Cancellation if paths from A_0 touch.
     *
     * @return
     */
    public boolean gammaHasConnectedRootComponents() {

        Multiset<GammaTuple> reachableVertices = HashMultiset.create();
        int numberOfRootTuples = 0;
        
        ConnectivityInspector ci = new ConnectivityInspector(gammaGraph);

        
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Gamma graph: " + gammaGraph+"\n\n");
        
        for (GammaTuple tuple : gamma.getGammaTuples()) {

            // Check if we have a root tuple:
            if (tuple.getConcept().equals(this.startConcept)) {

                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.err.println("Root tuple:    " + tuple);
                numberOfRootTuples++;

                // Get connected component containing this root tuple:
                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.err.println("  Current root tuple: " + tuple);
                
                reachableVertices.addAll(ci.connectedSetOf(tuple));
                    System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                    System.err.println("    Reachable tuples: " + ci.connectedSetOf(tuple));
            } else {
                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.err.println("No root tuple: " + tuple);
            }
        }

        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Number of root tuples in Gamma: " + numberOfRootTuples);
        for (Multiset.Entry<GammaTuple> tupleEntry : reachableVertices.entrySet()) {
            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
            System.err.println("Entry: " + tupleEntry);
        }

        // Now we have added all vertices on the paths from root tuples.
        // Some vertices might have been visited more than once if paths
        // from different root tuples touched. Check for this:
        for (Multiset.Entry<GammaTuple> tupleEntry : reachableVertices.entrySet()) {
            if (tupleEntry.getCount() > 1) {
                return true;
            }
        }

        return false;
    }

    /**
     * Answers 'true' if this Gamma gammaGraph contains a cycle. For the time
     * being, just call jgrapht routine.
     *
     * @return
     */
    public boolean graphIsCyclic() {
        CycleDetector cd = new CycleDetector(gammaGraph);
        return cd.detectCycles();
    }

    public static class RoleEdge<V> extends DefaultEdge {

        private V v1;
        private V v2;
        private OWLObjectPropertyExpression role;

        public RoleEdge(V v1, V v2, OWLObjectPropertyExpression role) {
            this.v1 = v1;
            this.v2 = v2;
            this.role = role;
        }

        public V getV1() {
            return v1;
        }

        public V getV2() {
            return v2;
        }

        public OWLObjectPropertyExpression getRole() {
            return role;
        }

        @Override
        public String toString() {
            return "[ edge " + role.toString() + " from " + v1 + " to " + v2 + "]";
        }
    }

}
