/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.omegaconstruction;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.context.OmegaContext;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogEntry;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.TimeUnit;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class OmegaSaturator {

    private Gamma gamma;
    private Omega omega;
    private Multimap<OWLClassExpression, GammaTuple> conceptToExGammaTuple;
    private Multimap<OWLClassExpression, OmegaTuple> conceptToOmegaTuple;
    private Set<OmegaTuple> queue, processed;
    private OmegaContext omegaContext;
    private DirectedGraph<OmegaTuple, DefaultEdge> omegaGraph;
    private Set<Edge> edges;
    private DetailsLogEntry logEntry;
    private OWLClass startConcept;
    private Context context;
    private boolean norewrite;

    // omegaToDatalog stuff:
    private final Map<OmegaTuple, Set<WitnessListPlusChosenSuccessor>> omegaTupleToSetOfListOfWitnesses = new HashMap<>();

    public Map<OmegaTuple, Set<WitnessListPlusChosenSuccessor>> getOmegaTupleToSetOfListOfWitnessesPlusSucc() {
        return omegaTupleToSetOfListOfWitnesses;
    }

    public Multimap<OWLClassExpression, OmegaTuple> getConceptToOmegaTuple() {
        return conceptToOmegaTuple;
    }

    public Omega getOmega() {
        return omega;
    }

    public OmegaSaturator(Gamma gamma, DetailsLogEntry logEntry, OWLClass startConcept, boolean norewrite, Context context) {
        this.context = context;
        this.norewrite = norewrite;
        this.logEntry = logEntry;
        this.gamma = gamma;
        this.startConcept = startConcept;
        this.omegaContext = context.getOmegaContext();
        initConceptToGammaTupleMappings();
        omegaGraph = new DefaultDirectedGraph<>(DefaultEdge.class);
        edges = new HashSet<>();
        //constructMCache = new HashMap<>();
    }

    public OmegaSaturator(Gamma gamma, OWLClass startConcept, Context context) {
        this(gamma, new DetailsLogEntry(), startConcept, true, context);
    }

    private void initConceptToGammaTupleMappings() {
        conceptToExGammaTuple = HashMultimap.create();
        for (GammaTuple tuple : gamma.getGammaTuples()) {
            for (GammaTuple otherTuple : gamma.getGammaTuples()) {
                for (OWLClassExpression concept : otherTuple.getSetOfConcepts()) {
                    if (concept instanceof OWLObjectSomeValuesFrom) {
                        OWLObjectSomeValuesFrom existential = (OWLObjectSomeValuesFrom) concept;
                        if (existential.getFiller().equals(tuple.getConcept())) {
                            conceptToExGammaTuple.put(tuple.getConcept(), otherTuple);
                        }
                    }
                }
            }
        }
    }

    public DirectedGraph<OmegaTuple, DefaultEdge> getOmegaGraph() {
        return omegaGraph;
    }

    /**
     * Generate the Omega tuples to start the saturation with. UPDATE: Here we
     * need to add an Omega tuple for EVERY Gamma tuple, not only for those
     * without existentials: We have to cater for the case that we lose all
     * existentials in a subsetting step!
     */
    private void generateOmegaZero() {
        queue = new HashSet<>();
        processed = new HashSet<>();
        conceptToOmegaTuple = HashMultimap.create();
        for (GammaTuple tuple : gamma.getGammaTuples()) {

            SortedSet<OWLClassExpression> sSetWithoutExistentials
                    = new TreeSet<>(context.getOriginalOnt().getComparator());
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLClass) {
                    sSetWithoutExistentials.add(concept);
                }
            }

            Set<OWLClassExpression> consequence
                    = omegaContext.closeUnderConsequence(sSetWithoutExistentials);
            OmegaTuple omegaTuple = new OmegaTuple(tuple.getConcept(), sSetWithoutExistentials, consequence, null, null);

            conceptToOmegaTuple.put(omegaTuple.getConcept(), omegaTuple);
            queue.add(omegaTuple);
            omegaGraph.addVertex(omegaTuple);
        }
    }

    public void saturate() {
        saturate(Long.MAX_VALUE);
    }

    /**
     * This implements the Omega rule to saturate the set Omega WITHOUT taking
     * subsets of the existentials in S, as we seem to have proofs for this
     * soon.
     *
     * @param timeoutPeriod
     */
    public void saturate(long timeoutPeriod) {

        if (timeoutPeriod < 0) {
            logEntry.setOmegaTimeout(true);
            System.out.println("Timed out, Omega phase!\n");
            return;
        }
        
        long timeoutBegin = System.currentTimeMillis();
        long usedUntilNow;
        boolean isTimeout;

        generateOmegaZero();

        // dumping for debugging OOMem Exception:
        long startTime = System.currentTimeMillis();
        long loopTime = System.currentTimeMillis();
        long numberOfLoops = 0;
        //

        // Do the following as long as we have unprocessed tuples in the pipeline
        // (and the time didn't run out):
        while (!queue.isEmpty()) {

            usedUntilNow = System.currentTimeMillis() - timeoutBegin;
            isTimeout = (usedUntilNow >= timeoutPeriod);
            if (isTimeout) {
                logEntry.setOmegaTimeout(true);
                System.out.println("Timed out, Omega phase!\n");
                return;
            }

            // Memleak debugging:
            final boolean debug = false;
            if (debug) {
                printMemoryDetails(loopTime, startTime, numberOfLoops);
            }

            // Normal taking from queue: Just any tuple.
            OmegaTuple currentOmegaTuple = queue.iterator().next();

            queue.remove(currentOmegaTuple);
            processed.add(currentOmegaTuple);

            if (debug) {
                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.err.println("Number of Gamma tuples touched for this Omega tuple: "
                        + conceptToExGammaTuple.get(currentOmegaTuple.getConcept()).size());
            }

            // Another great attempt to speed things up: 
            // Check for Omega cycles OFTEN! It's not too expensive!
            OmegaLoopChecker omegaLoopChecker = new OmegaLoopChecker(startConcept, omegaGraph, this);

            // Only check for Omega loops here, don't calculate reachable component!
            // If we find an Omega cycle, it won't be undone in the future.
            // Just return, and let GammaAndOmegaCaller do the loop checking 
            // once again (for the moment; TODO!).
            boolean omegaLoopFree = omegaLoopChecker.checkOmegaLoopFree();
            if (!omegaLoopFree) {
//                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                System.err.println("OMEGA IS LOOPING!");
                this.omega = new Omega(processed);
                return;
            }
            //

            // If we have Gamma tuples looking for a witness like currentTuple
            // to materialise into Omega tuples...
            for (GammaTuple gammaTupleLookingForWitness : conceptToExGammaTuple.get(currentOmegaTuple.getConcept())) {

                usedUntilNow = System.currentTimeMillis() - timeoutBegin;
                isTimeout = (usedUntilNow >= timeoutPeriod);
                if (isTimeout) {
                    logEntry.setOmegaTimeout(true);
                    System.out.println("Timed out, Omega phase!\n");
                    return;
                }

                // Extract the concept names and existentials from S:
                List<OWLObjectSomeValuesFrom> listOfAllExistentials = new ArrayList<>();
                Set<OWLClass> conceptNames = new HashSet<>();
                extractCNsAndExistentials(gammaTupleLookingForWitness, listOfAllExistentials, conceptNames);

                // For all existentials in S_t: Check for Omega witnesses and
                // subroles. Will continue for-loop if one witness set is empty!
                List<Set<OmegaTuple>> listOfOmegaWitnessSets = new ArrayList<>();
                List<Set<OWLObjectPropertyExpression>> listOfSubroleSets = new ArrayList<>();
                if (!fillAllOmegaWitnessesAndSubroles(listOfAllExistentials,
                        listOfOmegaWitnessSets, listOfSubroleSets)) {
                    continue;
                }

                // Now we have two lists of sets: listOfOmegaWitnessSets and listOfSubroles.
                // We have to combine this into all possible combinations,
                // producing a list of sets of OmegaTupleSubrolePairs.
                List<Set<OmegaTupleRolePair>> listOfOmegaTupleRolePairSets
                        = new ArrayList<>();

                for (int position = 0; position < listOfAllExistentials.size(); position++) {
                    Set<OmegaTupleRolePair> setOfPairsForOnePosition = new HashSet<>();
                    for (OmegaTuple oneWitness : listOfOmegaWitnessSets.get(position)) {
                        for (OWLObjectPropertyExpression oneRole : listOfSubroleSets.get(position)) {
                            OmegaTupleRolePair omegaTupleRolePair = new OmegaTupleRolePair(oneWitness, oneRole);
                            setOfPairsForOnePosition.add(omegaTupleRolePair);
                        }
                    }
                    listOfOmegaTupleRolePairSets.add(setOfPairsForOnePosition);
                }

                // For every existential, we might have several possible Omega 
                // witnesses and several possible subroles. 
                // Now we want to create a new Omega tuple for every 
                // possible combination of pairs (Omega witnesses, subrole of r_i)!
                Set<List<OmegaTupleRolePair>> allPossibleOmegaRolePairsForAllExistentials
                        = Sets.cartesianProduct(listOfOmegaTupleRolePairSets);

                for (List<OmegaTupleRolePair> onePairForEveryExistential : allPossibleOmegaRolePairsForAllExistentials) {

                    // This seems to be a valid optimisation: Check that every 
                    // list of Omega tuples actually contains currentTuple,
                    // as otherwise we did all of this before!
                    boolean found = false;
                    for (OmegaTupleRolePair singlePair : onePairForEveryExistential) {
                        if (singlePair.getOmegaTuple().equals(currentOmegaTuple)) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        continue;
                    }

                    //
                    Set<OWLClassExpression> conBegin = new HashSet<>();

                    // Initialise con/xcon of the new Omega tuples by adding all
                    // CNs of S of the template Gamma tuple.
                    conBegin.addAll(conceptNames);

                    // For all existentials in the subset: 
                    // Add to conBegin the \exists r.(con_t) for the Omega 
                    // tuple t used as witness for this existential and
                    // the role r.
                    for (int i = 0; i < listOfAllExistentials.size(); i++) {
                        addExistsRConT(onePairForEveryExistential.get(i).getOmegaTuple().getCon(),
                                onePairForEveryExistential.get(i).getRole(), conBegin);
                    }

                    Set<OWLClassExpression> conClosedUnderConsequence
                            = new HashSet<>(omegaContext.closeUnderConsequence(conBegin));

//                    // Now perform the famous check whether C_t is actually
//                    // in con_t: 
//                    // (Otherwise skip all the rest and start over again.)
                    if (!conClosedUnderConsequence.contains(gammaTupleLookingForWitness.getConcept())) {
//                        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                        System.err.println("Famous test failed (C_t not in con_t)!");
                        continue;
                    }
                    // For all existentials, produce different Omega tuples by 
                    // choosing different existentials as the chosen successor.
                    for (int chosenSuccessor = 0; chosenSuccessor < listOfAllExistentials.size(); chosenSuccessor++) {

                        // TODO: If chosenSuccessor points to currentOmegaTuple, we do not
                        // need to proceed ?!!
                        // Start xcon with the CNs of S:
                        Set<OWLClassExpression> xcon = new HashSet<OWLClassExpression>(conceptNames);

                        // Add to xcon the \exists r.(xcon_t) for the Omega 
                        // tuple t used as witness for this existential and the subrole r.
                        // Add the \exists r.(con_t') for the other Omega tuples t'
                        for (int otherSuccessor = 0; otherSuccessor < listOfAllExistentials.size(); otherSuccessor++) {
                            if (chosenSuccessor == otherSuccessor) {
                                addExistsRConT(onePairForEveryExistential.get(otherSuccessor).getOmegaTuple().getXCon(),
                                        onePairForEveryExistential.get(otherSuccessor).getRole(), xcon);
                            } else {
                                addExistsRConT(onePairForEveryExistential.get(otherSuccessor).getOmegaTuple().getCon(),
                                        onePairForEveryExistential.get(otherSuccessor).getRole(), xcon);
                            }
                        }

                        // Create one Omega tuple for the given Gamma tuple:
                        // For the chosen successor, we use the subrole s from the
                        // pair at position chosenSucc. The concept D is taken from
                        // this pair's OmegaTuple. We can only create the new
                        // OmegaTuple if \exists s.D is in the closure.
                        OWLObjectSomeValuesFrom existsSD = OWLManager.getOWLDataFactory().getOWLObjectSomeValuesFrom(
                                onePairForEveryExistential.get(chosenSuccessor).getRole(),
                                onePairForEveryExistential.get(chosenSuccessor).getOmegaTuple().getConcept());
                        OmegaTuple newOmegaTuple = new OmegaTuple(
                                gammaTupleLookingForWitness.getConcept(),
                                gammaTupleLookingForWitness.getSetOfConcepts(),
                                conClosedUnderConsequence,
                                listOfAllExistentials.get(chosenSuccessor),
                                omegaContext.closeUnderConsequence(xcon)
                        );

                        //
                        // !!WAS BELOW!!
                        //
                        // If the generated Omega tuple has already been created before,
                        // don't put it into the queue.
                        //if (processed.contains(newOmegaTuple) || queue.contains(newOmegaTuple)) {
                        //    continue;
                        // }
                        if (!norewrite) {

                            // Here is the place to update the Map
                            //   OmegaTuple -> Set of Lists of witness OmegaTuples.
                            // First, construct a list out of the current witnesses:
                            WitnessListPlusChosenSuccessor witnessListSucc
                                    = new WitnessListPlusChosenSuccessor(onePairForEveryExistential, chosenSuccessor);

                            // Then check if we have an identical Witness already 
                            // in the Map. Add the current list of witnesses in that case:
                            if (omegaTupleToSetOfListOfWitnesses.containsKey(newOmegaTuple)) {
                                Set<WitnessListPlusChosenSuccessor> setOfWitnessLists
                                        = omegaTupleToSetOfListOfWitnesses.get(newOmegaTuple);
                                setOfWitnessLists.add(witnessListSucc);
                            } // Else: this OmegaTuple is really new. Add a new Map entry
                            // for it using the current list of witnesses.
                            else {
                                Set<WitnessListPlusChosenSuccessor> setOfListsOfWitnesses = new HashSet<>();
                                setOfListsOfWitnesses.add(witnessListSucc);
                                omegaTupleToSetOfListOfWitnesses.put(newOmegaTuple, setOfListsOfWitnesses);
                            }
                        }//
                        //

                        // Update the Omega graph. The new tuple is added
                        // as a vertex:
                        omegaGraph.addVertex(newOmegaTuple);
                        // Its chosen successor is added as a successor in the graph.
                        // HACK: we add edges later to the graph because 
                        //       on the fly adding is slow with JGraphT.
                        edges.add(new Edge(newOmegaTuple, onePairForEveryExistential.get(chosenSuccessor).getOmegaTuple()));
                        // Update the mapping using this Omega tuple.
                        conceptToOmegaTuple.put(newOmegaTuple.getConcept(), newOmegaTuple);

                        //
                        // !!WAS HERE!!
                        //
                        if (processed.contains(newOmegaTuple) || queue.contains(newOmegaTuple)) {
                            continue;
                        }
                        queue.add(newOmegaTuple);
                    }
                }
            }

            //
            // For optimisation 2: we need a filled Omega graph, pronto!
            for (Edge edge : edges) {
                omegaGraph.addEdge(edge.getSource(), edge.getTarget());
            }
            //

        }

        logEntry.setOmegaTimeout(false);

        // TODO: Maybe this in not needed (did it in inner loop)?
        for (Edge edge : edges) {
            omegaGraph.addEdge(edge.getSource(), edge.getTarget());
        }
        this.omega = new Omega(processed);

    }

    private void printMemoryDetails(long loopTime, long startTime, long numberOfLoops) {
        long loopDelta = (System.currentTimeMillis() - loopTime);
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Time spent processing last Omega tuple: " + loopDelta + " ms.");
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Avegare time for one Omega tuple:      " + ((System.currentTimeMillis() - startTime) / (++numberOfLoops)) + " ms.");
        loopTime = System.currentTimeMillis();
        this.dumpInternals();
        long millis = System.currentTimeMillis() - startTime;
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.format("Time running:         %4d min, %2d sec.%n",
                TimeUnit.MILLISECONDS.toMinutes(millis),
                TimeUnit.MILLISECONDS.toSeconds(millis)
                - TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(millis))
        );
        millis = Math.round((System.currentTimeMillis() - startTime) * (((double) queue.size()) / processed.size()));
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.format("Estimated time to go: %4d min, %2d sec.%n",
                TimeUnit.MILLISECONDS.toMinutes(millis),
                TimeUnit.MILLISECONDS.toSeconds(millis)
                - TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(millis))
        );
        this.dumpMemsize();
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println();
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Take one from queue.");
    }

    /**
     * For all existentials in S_t: Check for Omega witnesses!
     *
     * @param listOfExistentialsNeedingOmegaWitness
     * @param listOfOmegaWitnessSets
     * @return
     */
    private boolean fillAllOmegaWitnessesAndSubroles(
            List<OWLObjectSomeValuesFrom> listOfExistentialsNeedingOmegaWitness,
            List<Set<OmegaTuple>> listOfOmegaWitnessSets,
            List<Set<OWLObjectPropertyExpression>> listOfSubroleSets) {

        for (OWLObjectSomeValuesFrom existential : listOfExistentialsNeedingOmegaWitness) {

            // If we do not find an Omega witness for one existential
            // in S_t, then we cannot produce any new Omega tuples.
            //
            // When we still did the subset-building of ex. in S:
            // We still have to check the other subsets of existentials,
            // as some might miss the current existential w/o witness!
            // TODO: Room for optimisation, check only the subsets
            //       left w/o the current ex.
            if (!conceptToOmegaTuple.containsKey(existential.getFiller())) {
                return false;
            }

            Set<OmegaTuple> setOfOmegaWitnesses = new HashSet<>();
            setOfOmegaWitnesses.addAll(conceptToOmegaTuple.get(existential.getFiller()));
            listOfOmegaWitnessSets.add(setOfOmegaWitnesses);

            Set<OWLObjectPropertyExpression> setOfSubroles = new HashSet<>();
            setOfSubroles.addAll(context.getReasoning().getEntailingRoles(existential.getProperty()));
            listOfSubroleSets.add(setOfSubroles);
        }

        return true;
    }

    /**
     * Extract the concept names from S (these always stay in S_t), and the
     * existentials (these will be subset-ted):
     */
    private void extractCNsAndExistentials(GammaTuple gammaTupleLookingForWitness,
            List<OWLObjectSomeValuesFrom> listOfAllExistentials, Set<OWLClass> conceptNames) {
        for (OWLClassExpression concept : gammaTupleLookingForWitness.getSetOfConcepts()) {
            if (concept instanceof OWLObjectSomeValuesFrom) {
                OWLObjectSomeValuesFrom existential = (OWLObjectSomeValuesFrom) concept;
                listOfAllExistentials.add(existential);
            } else {
                OWLClass conceptName = (OWLClass) concept;
                conceptNames.add(conceptName);
            }
        }
    }

    /**
     * Add \exists relevantRole.C for all C in cxon.
     *
     * @param cxon The consequences to add via relevantRole.
     * @param relevantRole
     * @param setM This set will be added to!
     */
    private void addExistsRConT(Set<OWLClassExpression> cxon,
            OWLObjectPropertyExpression relevantRole, Set<OWLClassExpression> setM) {

        // If we have to construct \exists r.--, just add \top if it is in sub(T).
        if (cxon == null) {
            if (context.getOriginalOnt().getSubconcepts().contains(context.topConcept())) {
                setM.add(context.topConcept());
            }
            return;
        }
        for (OWLClassExpression g : cxon) {

            // Put all exists [superrole of r].G into con! This will also help
            // if exists r.G is not in sub(T)!
            for (OWLObjectPropertyExpression superrole : context.getReasoning().getEntailedRoles(relevantRole)) {
                OWLObjectSomeValuesFrom existsRG
                        = OWLManager.getOWLDataFactory().getOWLObjectSomeValuesFrom(superrole, g);
                if (omegaContext.getContext().getOriginalOnt().getSubconcepts().contains(existsRG)) {
                    setM.add(existsRG);
                }
            }
        }
    }

    public void dumpInternals() {
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Queue size: " + queue.size());
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Processed size: " + processed.size());
    }

    public void dumpMemsize() {

        int mb = 1024 * 1024;

        //Getting the runtime reference from system
        Runtime runtime = Runtime.getRuntime();

        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("##### Heap utilization statistics [MB] #####");
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Used Memory:"
                + (runtime.totalMemory() - runtime.freeMemory()) / mb);
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Free Memory:"
                + runtime.freeMemory() / mb);
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Total Memory:" + runtime.totalMemory() / mb);
        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
        System.err.println("Max Memory:" + runtime.maxMemory() / mb);
    }

    /**
     * Used for the jGraphT hack of adding edges later to gain speed.
     *
     * @param <OmegaTuple>
     */
    private class Edge {

        private final OmegaTuple source;
        private final OmegaTuple target;

        private Edge(OmegaTuple source, OmegaTuple target) {
            this.source = source;
            this.target = target;
        }

        public OmegaTuple getSource() {
            return this.source;
        }

        public OmegaTuple getTarget() {
            return this.target;
        }

        private volatile int hash;

        @Override
        public int hashCode() {
            if (hash == 0) {
                hash = 5;
                hash = 71 * hash + Objects.hashCode(this.source);
                hash = 71 * hash + Objects.hashCode(this.target);
            }
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final Edge other = (Edge) obj;
            if (!Objects.equals(this.source, other.source)) {
                return false;
            }
            if (!Objects.equals(this.target, other.target)) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "(" + source + " --> " + target + ")";
        }
    }

    public class WitnessListPlusChosenSuccessor {

        private final List<OmegaTupleRolePair> witnessList;
        private final int chosenSuccessor;

        public WitnessListPlusChosenSuccessor(List<OmegaTupleRolePair> witnessList, int chosenSuccessor) {
            this.witnessList = witnessList;
            this.chosenSuccessor = chosenSuccessor;
        }

        public List<OmegaTupleRolePair> getWitnessList() {
            return witnessList;
        }

        public int getChosenSuccessor() {
            return chosenSuccessor;
        }

        private volatile int hash;

        @Override
        public int hashCode() {
            if (hash == 0) {
                hash = 3;
                hash = 17 * hash + Objects.hashCode(this.witnessList);
                hash = 17 * hash + this.chosenSuccessor;
            }
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final WitnessListPlusChosenSuccessor other = (WitnessListPlusChosenSuccessor) obj;
            if (!Objects.equals(this.witnessList, other.witnessList)) {
                return false;
            }
            if (this.chosenSuccessor != other.chosenSuccessor) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "WitnessListPlusChosenSuccessor{" + "witnessList=" + witnessList + ", chosenSuccessor=" + chosenSuccessor + "}\n";
        }

    }

    /**
     * This holds a possible combination of a subrole and an OmegaTuple.
     */
    public class OmegaTupleRolePair {

        private final OmegaTuple omegaTuple;
        private final OWLObjectPropertyExpression role;

        public OmegaTupleRolePair(OmegaTuple omegaTuple, OWLObjectPropertyExpression role) {
            this.omegaTuple = omegaTuple;
            this.role = role;
        }

        public OmegaTuple getOmegaTuple() {
            return omegaTuple;
        }

        public OWLObjectPropertyExpression getRole() {
            return role;
        }

        private volatile int hash;

        @Override
        public int hashCode() {
            if (hash == 0) {
                hash = 3;
                hash = 79 * hash + Objects.hashCode(this.omegaTuple);
                hash = 79 * hash + Objects.hashCode(this.role);
            }
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final OmegaTupleRolePair other = (OmegaTupleRolePair) obj;
            if (!Objects.equals(this.omegaTuple, other.omegaTuple)) {
                return false;
            }
            if (!Objects.equals(this.role, other.role)) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "(" + omegaTuple + ", " + role + ")";
        }
    }

    /**
     * Add \exists [subrole].C for all C in cxon and subroles of relevantRole.
     *
     * @param cxon The consequences to add via relevantRole.
     * @param relevantRole
     * @param setM This set will be added to!
     */
    @Deprecated
    private void addExistsSConTOLD(Set<OWLClassExpression> cxon,
            OWLObjectPropertyExpression relevantRole, Set<OWLClassExpression> setM) {

        // If we have to construct \exists r.--, just add \top if it is in sub(T).
        if (cxon == null) {
            if (context.getOriginalOnt().getSubconcepts().contains(context.topConcept())) {
                setM.add(context.topConcept());
            }
            return;
        }
        for (OWLClassExpression g : cxon) {
            for (OWLObjectPropertyExpression subrole
                    : context.getReasoning().getEntailingRoles(relevantRole)) {
                OWLObjectSomeValuesFrom existsSG
                        = OWLManager.getOWLDataFactory().getOWLObjectSomeValuesFrom(subrole, g);
                if (omegaContext.getContext().getOriginalOnt().getSubconcepts().contains(existsSG)) {
                    setM.add(existsSG);
                }
            }
        }
    }

    /**
     * This implements the Omega rule to saturate the set Omega INCLUDING taking
     * subsets of the existentials of S. Due to a change in the proofs, this is
     * not called at the moment.
     *
     * @param timeoutPeriod
     * @return
     */
    /*
     @Deprecated
     public Omega saturateWithSubsets(long timeoutPeriod) {
     generateOmegaZero();

     boolean timeout = false;
     long timeoutEnd;
     long timeoutBegin = System.currentTimeMillis();

     // Do the following as long as we have unprocessed tuples in the pipeline
     // (and the time didn't run out):
     while (!queue.isEmpty() && !timeout) {
     OmegaTuple currentTuple = queue.iterator().next();
     queue.remove(currentTuple);
     processed.add(currentTuple);

     // If we have Gamma tuples looking for a witness like currentTuple
     // to materialise into Omega tuples...
     for (GammaTuple gammaTupleLookingForWitness : conceptToExGammaTuple.get(currentTuple.getConcept())) {

     List<OWLObjectSomeValuesFrom> listOfAllExistentials = new ArrayList<>();
     Set<OWLClass> conceptNames = new HashSet<>();

     // Extract the concept names from S (these always stay in S_t),
     // and the existentials (these will be subset-ted):
     extractCNsAndExistentials(gammaTupleLookingForWitness, listOfAllExistentials, conceptNames);

     // Now we consider all subsets of the set of existentials in S of our
     // Gamma tuple. There are 2^|ex| many!
     // Watch out! We start with 1 because we want at least one 
     // existential to stay in the set!!
     for (int binaryCounter = 1; binaryCounter < (int) Math.pow(2, listOfAllExistentials.size()); binaryCounter++) {

     List<OWLObjectSomeValuesFrom> subsetListOfExistentialsNeedingOmegaWitness = new ArrayList<>();
     List<Set<OmegaTuple>> listOfOmegaWitnessSets = new ArrayList<>();

     // Generate List of current existentials in S_t:
     generateListOfExistentials(listOfAllExistentials, binaryCounter, subsetListOfExistentialsNeedingOmegaWitness);

     // For all existentials now still in S_t: Check for Omega witnesses!
     if (!fillAllOmegaWitnessesAndSubroles(subsetListOfExistentialsNeedingOmegaWitness, listOfOmegaWitnessSets)) {
     continue;
     }

     // For every existential in the subset of existentials, we might have
     // several possible Omega witnesses. Now we want to create a new Omega
     // tuple for every possible combination of witnesses.
     Set<List<OmegaTuple>> childrenOmegas = Sets.cartesianProduct(listOfOmegaWitnessSets);

     for (List<OmegaTuple> omegaList : childrenOmegas) {
     Set<OWLClassExpression> conBegin = new HashSet<>();

     // Initialise con/xcon of the new Omega tuples by adding all
     // CNs of S of the template Gamma tuple.
     conBegin.addAll(conceptNames);

     // For all existentials in the subset: 
     // Add to con the \exists s.(con_t) for the Omega tuple t 
     // used as witness for this existential.
     for (int i = 0; i < subsetListOfExistentialsNeedingOmegaWitness.size(); i++) {
     OWLObjectPropertyExpression role = subsetListOfExistentialsNeedingOmegaWitness.get(i).getProperty();
     addExistsSConT(omegaList.get(i).getCon(), role, conBegin);
     }

     Set<OWLClassExpression> conClosedUnderConsequence
     = new HashSet<>(omegaContext.closeUnderConsequence(conBegin));

     // Now perform the famous check whether C_t is actually
     // in con_t: 
     // (Otherwise skip all the rest and start over again.)
     if (!conClosedUnderConsequence.contains(gammaTupleLookingForWitness.getConcept())) {
     continue;
     }

     // For all existentials in the subset, produce Omega tuples
     // by choosing one existential as the chosen successor.
     for (int chosenSuccessor = 0; chosenSuccessor < subsetListOfExistentialsNeedingOmegaWitness.size(); chosenSuccessor++) {

     // Start xcon with the CNs of S:
     Set<OWLClassExpression> xcon = new HashSet<OWLClassExpression>(conceptNames);

     // Add to xcon the \exists s.(xcon_t) for the Omega 
     // tuple t used as witness for this existential;
     // add the \exists s.(con_t') for the other Omega tuples t'
     for (int otherSuccessor = 0; otherSuccessor < subsetListOfExistentialsNeedingOmegaWitness.size(); otherSuccessor++) {
     OWLObjectPropertyExpression role = subsetListOfExistentialsNeedingOmegaWitness.get(otherSuccessor).getProperty();
     if (chosenSuccessor == otherSuccessor) {
     addExistsSConT(omegaList.get(otherSuccessor).getXCon(), role, xcon);
     } else {
     addExistsSConT(omegaList.get(otherSuccessor).getCon(), role, xcon);
     }
     }

     OmegaTuple newTuple = new OmegaTuple();
     newTuple.setConcept(gammaTupleLookingForWitness.getConcept());

     // Create the new S set for the new tuple:
     SortedSet<OWLClassExpression> newSetOfConcepts = new TreeSet<>(omegaContext.getContext().getOriginalOnt().getComparator());
     newSetOfConcepts.addAll(conceptNames);
     newSetOfConcepts.addAll(subsetListOfExistentialsNeedingOmegaWitness);
     newTuple.setSetOfConcepts(newSetOfConcepts);

     newTuple.setCon(conClosedUnderConsequence);
     newTuple.setSuccessor(subsetListOfExistentialsNeedingOmegaWitness.get(chosenSuccessor));
     newTuple.setXcon(omegaContext.closeUnderConsequence(xcon));

     // If the generated Omega tuple has already been created before,
     // skip the rest.
     if (processed.contains(newTuple) || queue.contains(newTuple)) {
     continue;
     }

     // Update the mapping using this Omega tuple.
     conceptToOmegaTuple.put(newTuple.getConcept(), newTuple);

     queue.add(newTuple);

     // Update the Omega graph. The new tuple is added
     // as a vertex, and its chosen successor is added
     // as a successor in the graph.
     omegaGraph.addVertex(newTuple);
     // HACK: we add edges later to the graph because 
     // on the fly adding is slow with JGraphT
     edges.add(new Edge(newTuple, omegaList.get(chosenSuccessor)));
     }
     }
     }
     }
     timeoutEnd = System.currentTimeMillis();
     timeout = (timeoutEnd - timeoutBegin > timeoutPeriod);
     }
     logEntry.setOmegaTimeout(timeout);
     if (timeout) {
     System.err.println("Timed out, Omega phase!\n");
     }
     for (Edge edge : edges) {
     omegaGraph.addEdge(edge.getSource(), edge.getTarget());
     }
     return new Omega(processed);
     }
     */
    /**
     * Generate List of current existentials in S_t:
     *
     * @param listOfAllExistentials
     * @param binaryCounter
     * @param subsetListOfExistentialsNeedingOmegaWitness
     */
    /*    @Deprecated
     private void generateListOfExistentials(List<OWLObjectSomeValuesFrom> listOfAllExistentials,
     int binaryCounter,
     List<OWLObjectSomeValuesFrom> subsetListOfExistentialsNeedingOmegaWitness) {

     for (int position = 0; position < listOfAllExistentials.size(); position++) {
     // Is this existential in the current subset?
     if ((((int) Math.pow(2, position)) & binaryCounter) != 0) {
     subsetListOfExistentialsNeedingOmegaWitness.add(listOfAllExistentials.get(position));
     }
     }
     }
     */
}
