/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.runnables;

import de.unibremen.informatik.tdki.grind.control.GammaAndOmegaCaller;
import de.unibremen.informatik.tdki.grind.statistics.OverviewLogger;
import java.io.FileNotFoundException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import net.sourceforge.argparse4j.ArgumentParsers;
import net.sourceforge.argparse4j.impl.Arguments;
import net.sourceforge.argparse4j.inf.ArgumentParser;
import net.sourceforge.argparse4j.inf.ArgumentParserException;
import net.sourceforge.argparse4j.inf.Namespace;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;

/**
 * This is the class to be called from the command line. It distributes work to
 * RunManyOntologies and RunSingleConceptname according to command line
 * arguments.
 *
 * @author hansen
 */
public class MainRunnable {

    final static String VERSION = "1.0 (2015-04-24)";

    /**
     * Main method called from command line using at least the ontology and a
     * query as arguments.
     *
     * @param args
     * @throws org.semanticweb.owlapi.model.OWLOntologyCreationException
     * @throws java.io.FileNotFoundException
     */
    //Example:
    //   $ grind MOHSE.owl Bacillus
    // 
    // Output will be a rewriting of concept Bacillus in Datalog syntax
    // to stdout, e.g.
    // Q(?0)  <-  Bacillus(?0)
    // Q(?0)  <-  BacillusAnthracis(?0)
    // Q(?0)  <-  BacillusCereus(?0)
    public static void main(String[] args) throws OWLOntologyCreationException,
            FileNotFoundException,
            Exception {

        // Parsing of command line arguments is done using argparse4j. 
        // Argument parsing in java is a royal pain, let's hope this
        // library stays around for some time.
        ArgumentParser parser = ArgumentParsers.newArgumentParser("grind")
                //.defaultHelp(true)
                .description("Rewrite the concept to non-recursive datalog with "
                        + "respect to the ontology. Result is written to stdout.");
        parser.addArgument("-v", "--verbose")
                .dest("verbose")
                .action(Arguments.storeConst())
                .setConst(true)
                .setDefault(false)
                .help("dump internals to stdout");
        parser.addArgument("-l", "--logging")
                .dest("logging")
                .action(Arguments.storeConst())
                .setConst(true)
                .setDefault(false)
                .help("dump some more internals to ./grind-logs/ (only with --all)");
        parser.addArgument("-a", "--all")
                .dest("all")
                .action(Arguments.storeConst())
                .setConst(true)
                .setDefault(false)
                .help("rewrite all concept names in ontology");
        parser.addArgument("-n", "--norewrite")
                .dest("norewrite")
                .action(Arguments.storeConst())
                .setConst(true)
                .setDefault(false)
                .help("only check FO-rewritability, do not construct an actual rewriting");
//        parser.addArgument("-t", "--timeout-gamma")
//                .dest("gammaTimeout")
//                .action(Arguments.storeConst())
//                .setConst(true)
//                .setDefault(5_000)
//                .help("Timeout (in ms) for Gamma phase. Defaults to 5 s.");
//        parser.addArgument("-T", "--timeout-omega")
//                .dest("omegaTimeout")
//                .action(Arguments.storeConst())
//                .setConst(true)
//                .setDefault(5_000)
//                .help("Timeout (in ms) for Omega phase. Defaults to 15 min.");
        parser.addArgument("ontology")
                .help("an OWL ontology (e.g. 'MOHSE.owl')");
        parser.addArgument("concept")
                .nargs("*")
                .help("one or more concept queries (e.g. 'Bacillus Fungus Virus')");
        parser.epilog("Grind -- Gracefully Rewriting Into Non-recursive Datalog.\n"
                + "" + VERSION + ".\n"
                                + "Copyright 2014, 2015 Peter Hansen and İnanç Seylan.\n"
                                + "University of Bremen, AG TdKI.\n"
                                + "Homepage: http://www.informatik.uni-bremen.de/~hansen/grind\n"
                + "This software is released under GPL 3 licence.\n"
                + "\n"
                + "Supports the following types of axioms in an ontology:\n"
                + " * <SubClassOf>\n"
                + " * <EquivalentClasses>\n"
                + " * <SubObjectPropertyOf>\n"
                + " * <EquivalentObjectProperties>\n"
                + " * <ObjectPropertyDomain>\n"
                + "\n"
                + "and the following concept constructors:\n"
                + " * <ObjectIntersectionOf>\n"
                + " * <ObjectSomeValuesFrom>\n"
        );
        parser.version(VERSION);
        Namespace ns = null;
        try {
            ns = parser.parseArgs(args);
        } catch (ArgumentParserException e) {
            parser.handleError(e);
            System.exit(-1);
        }

        // We won't write log files without going over all concepts:
        if (ns.getBoolean("logging") && !ns.getBoolean("all")) {
            System.err.println("grind: Use --logging only together with --all. Aborting.");
            System.exit(-1);
        }

        GammaAndOmegaCaller caller = new GammaAndOmegaCaller(ns.getString("ontology"),
                ns.getBoolean("logging"), ns.getBoolean("verbose"), ns.getBoolean("norewrite"));

        if (ns.getBoolean("logging")) {
            Date dNow = new Date();
            SimpleDateFormat ft = new SimpleDateFormat("yyyyMMdd-HHmmss");
            System.out.println("Current Date: " + ft.format(dNow));
            String date = ft.format(dNow);
            OverviewLogger.initialize(ns.getString("ontology"), date, ns.getBoolean("logging"));
        }

        List<String> listOfConcepts = (List<String>) ns.get("concept");

        // The user can choose between rewriting all concept names in ont (--a):
        if (ns.getBoolean("all")) {
            if (!listOfConcepts.isEmpty()) {
                System.err.println("grind: Option --all cannot take concept name parameters. Aborting.");
                System.exit(-1);
            }
            caller.runAllConceptnames();
            if (ns.getBoolean("logging")) {
                OverviewLogger.getInstance().close();
            }
            // Return a success here:
            System.exit(0);
        }

        // Or they can state a list of CNs to rewrite:
        // Do a switch-case on the length of the list:
        switch (listOfConcepts.size()) {

            // No concept names provided:
            case 0:
                System.err.println("grind: Missing concept names to rewrite. Aborting.");
                System.exit(-1);

            // If we only state one CN as argument, we get a sensible return 
            // value to the shell: 0 for FO-rewr, 1 for Not FO-rewr, -1 for timeout:
            case 1:
                int retval = caller.runSingleConceptname(listOfConcepts.get(0));
                if (ns.getBoolean("logging")) {
                    OverviewLogger.getInstance().close();
                }
                System.exit(retval);

            default:
                for (String concept : listOfConcepts) {
                    System.out.println("\n" + concept + ":\n");
                    caller.runSingleConceptname(concept);
                    System.out.println();
                }
                if (ns.getBoolean("logging")) {
                    OverviewLogger.getInstance().close();
                }
                System.exit(0);
        }

        // Should not be reached:
        System.exit(-1);
    }

}
