/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.statistics;

import java.io.*;
import java.util.Map;
import java.util.SortedMap;
import java.util.StringTokenizer;
import java.util.TreeMap;

/**
 * This serves to collect data written to <detailsDir>/<OntoName>.txt by 
 * StatisticsLogger. The txt file is parsed, and for each concept, if
 * we have no timeout, statistical info (logarithmic Gamma size, etc.)
 * is written to 
 *   <detailsDir>/<OntoName>_gamma_size.dat 
 *   <detailsDir>/<OntoName>_lhs_size.dat
 *   <detailsDir>/<OntoName>_rhs_max_size.dat
 *   <detailsDir>/<OntoName>_rhs_sum_size.dat
 *   <detailsDir>/<OntoName>_rhs_avg_size.dat
 */
public class DetailsEvaluator {
        static String detailsDir = "./grind-logs/20150209-091720-seven-ontologies-GOGO-bugfix/";

        public static void main(String[] args) throws FileNotFoundException, IOException {
        
        String[] detailsFilenames = new String[]{
            // our original results
//            "./logs/ENVO.txt",
//            "./logs/FBbi.txt",
//            "./logs/MOHSE.txt",
//            "./logs/NBO.txt",
//            "./logs/SO.txt",
//            "./logs/XP.txt",
//            "./logs/not-galen.txt", 

            // REQUIEM's results
//            "./logs/requiem-ENVO-resaved-protege-402.txt",
//            "./logs/requiem-MOHSE.txt",
//            "./logs/requiem-NBO.txt",
//            "./logs/requiem-not-galen.txt", 
//            "./logs/requiem-XP.txt",
//            "./logs/requiem-SO.txt",
//            "./logs/requiem-FBbi.txt",

            // our original results, filtered to only CNs that REQ can process
//            "./logs/filtered-ENVO-resaved-protege-402.txt",
//            "./logs/filtered-MOHSE.txt",
//            "./logs/filtered-NBO.txt",
//            "./logs/filtered-not-galen.txt", 
//            "./logs/filtered-SO.txt",
//            "./logs/filtered-XP.txt",
//            "./logs/filtered-FBbi.txt",
            
            detailsDir + "ENVO-resaved-protege-402-details.txt",
            detailsDir + "FBbi-resaved-protege-402-details.txt",
            detailsDir + "MOHSE-details.txt",
            detailsDir + "NBO-details.txt",
            detailsDir + "not-galen-details.txt",
            detailsDir + "SO-from-obo-402-details.txt",
            detailsDir + "XP-details.txt",
        };
        for (String s : detailsFilenames) {
            generateStatistics(s);
        }
    }

    private static int log(int no) {
        double base = 2;
        return (int) (Math.log(no) / Math.log(base));
    }

    private static void safeAddToMap(int key, Map<Integer, Integer> map) {
        if (map.containsKey(key)) {
            int value = map.get(key);
            map.put(key, value + 1);
        } else {
            map.put(key, 1);
        }
    }

    private static void flushMap(SortedMap<Integer, Integer> map, String filename, int totalCNSize) throws FileNotFoundException {
        PrintWriter writer = null;
        try {
            writer = new PrintWriter(filename);
            for (int i = 0; i <= 13; i++) { // not until map.lastKey()
                writer.print((int) Math.pow(2, i));
                writer.print("\t");
                if (map.containsKey(i)) {
                    writer.println(((float) (100 * map.get(i))) / totalCNSize);
                } else {
                    writer.println(0);
                }
            }
        } finally {
            if (writer != null) {
                writer.close();
            }
        }
    }

    private static void flushMapLhs(SortedMap<Integer, Integer> map, String filename, int totalCNSize) throws FileNotFoundException {
        PrintWriter writer = null;
        try {
            writer = new PrintWriter(filename);
            for (int i = 0; i <= 6; i++) { // not until map.lastKey()
                writer.print((int) Math.pow(2, i));
                writer.print("\t");
                if (map.containsKey(i)) {
                    writer.println(((float) (100 * map.get(i))) / totalCNSize);
                } else {
                    writer.println(0);
                }
            }
        } finally {
            if (writer != null) {
                writer.close();
            }
        }
    }

    private static void generateStatistics(String source) throws FileNotFoundException, IOException {
        BufferedReader reader = new BufferedReader(new FileReader(source));
        String line;
        SortedMap<Integer, Integer> gammaSizeDistribution = new TreeMap<>();
        SortedMap<Integer, Integer> lhsSizeDistribution = new TreeMap<>();
        SortedMap<Integer, Integer> rhsMaxSizeDistribution = new TreeMap<>();
        SortedMap<Integer, Integer> rhsSizeSumDistribution = new TreeMap<>();
        SortedMap<Integer, Integer> rhsAvgSizeDistribution = new TreeMap<>();
        int totalCNSize = 0;
        while ((line = reader.readLine()) != null) {
            StringTokenizer tokenizer = new StringTokenizer(line, ",");
            boolean success = false;
            for (int i = 0; i < 9; ++i) { 
                String item = tokenizer.nextToken();
                // 0 is the concept name
                //   (For info about the entries, see LogEntry.java!)
                // 1 is "success", "gammaTimeout", or "omegaTimeout"
                if (i == 1 && item.equals("success")) {
                    success = true;
                    totalCNSize++;
                }
                // 2 is gammaSize
                if (i == 2 && success) {
                    int gammaSize = Integer.valueOf(item);
                    safeAddToMap(log(gammaSize), gammaSizeDistribution);
                }
                // 3 is 
                if (i == 3 && success) {
                    int noOfDifferentGammaLhs = Integer.valueOf(item);
                    safeAddToMap(log(noOfDifferentGammaLhs), lhsSizeDistribution);
                }
                if (i == 4 & success) {
                    int maxRhs = Integer.valueOf(item);
                    safeAddToMap(log(maxRhs), rhsMaxSizeDistribution);
                }
                if (i == 5 & success) {
                    int sumRhs = Integer.valueOf(item);
                    safeAddToMap(log(sumRhs), rhsSizeSumDistribution);
                }
                if (i == 6 & success) {
                    float avgRhs = Float.parseFloat(item);
                    safeAddToMap(log((int) avgRhs), rhsAvgSizeDistribution);
                }
            }
        }
        File f = new File(source);
        flushMap(gammaSizeDistribution, detailsDir + f.getName().replace("-details.txt", "_gamma_size.dat"), totalCNSize);
        flushMapLhs(lhsSizeDistribution, detailsDir + f.getName().replace("-details.txt", "_lhs_size.dat"), totalCNSize);

        // flush also directly into Latex directory
//        flushMap(gammaSizeDistribution, "/home/hansen/hansen-repo/IJCAI2015/plots/" + f.getName().replace(".txt", "_gamma_size.dat"), totalCNSize);
//        flushMapLhs(lhsSizeDistribution, "/home/hansen/hansen-repo/IJCAI2015/plots/" + f.getName().replace(".txt", "_lhs_size.dat"), totalCNSize);
  
        // ?
//        flushMap(rhsMaxSizeDistribution, "./logs/" + f.getName().replace(".txt", "_rhs_max_size.dat"), totalCNSize);
//        flushMap(rhsSizeSumDistribution, "./logs/" + f.getName().replace(".txt", "_rhs_sum_size.dat"), totalCNSize);
//        flushMap(rhsAvgSizeDistribution, "./logs/" + f.getName().replace(".txt", "_rhs_avg_size.dat"), totalCNSize);
    }
}
