/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.utilities;

import de.unibremen.informatik.tdki.grind.utilities.counting.ConceptNameCollectingVisitor;
import de.unibremen.informatik.tdki.grind.utilities.counting.OntologyStatistics;
import de.unibremen.informatik.tdki.grind.utilities.counting.SubconceptCollectingVisitor;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectIntersectionOf;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;

/**
 *
 * @author seylan
 */
public class ELUtils {
    
    public static Set<OWLClassExpression> getSubconcepts(OWLClassExpression e) {
        SubconceptCollectingVisitor v = new SubconceptCollectingVisitor();
        e.accept(v);
        return v.getConcepts();
    }
    
    public static Set<OWLClassExpression> getSubconcepts(Set<OWLSubClassOfAxiom> axioms) {
        Set<OWLClassExpression> result = new HashSet<>();
        for (OWLSubClassOfAxiom axiom : axioms) {
            result.addAll(getSubconcepts(axiom.getSubClass()));
            result.addAll(getSubconcepts(axiom.getSuperClass()));
        }
        return result;
    }
    
    public static Set<OWLClass> getConceptNames(Set<OWLSubClassOfAxiom> axioms) {
        ConceptNameCollectingVisitor v = new ConceptNameCollectingVisitor();
        for (OWLSubClassOfAxiom axiom : axioms) {
            axiom.getSubClass().accept(v);
            axiom.getSuperClass().accept(v);
        }
        return v.getConceptsNames();
    }
    
    public static Set<OWLObjectProperty> getRoleNames(Set<OWLSubClassOfAxiom> axioms) {
        Set<OWLObjectProperty> result = new HashSet<>();
        for (OWLClassExpression c : ELUtils.getSubconcepts(axioms)) {
            if (c instanceof OWLObjectSomeValuesFrom) {
                OWLObjectSomeValuesFrom exists = (OWLObjectSomeValuesFrom) c;
                result.add(exists.getProperty().asOWLObjectProperty());
            }
        }
        return result;
    }
    
    public static Set<OWLClassExpression> getTlcs(OWLClassExpression e) {
        return e.asConjunctSet();
    }
    
    @Deprecated
    public static boolean isExistentialRestriction(OWLClassExpression e) {
        return (e instanceof OWLObjectSomeValuesFrom);
    }
    
    @Deprecated
    public static boolean isConceptName(OWLClassExpression e) {
        return (e instanceof OWLClass);
    }
    
    public static int getDepth(OWLClassExpression concept) {
        if (concept instanceof OWLClass) {
            return 0;
        }
        if (concept instanceof OWLObjectIntersectionOf) {
            int maximum = 0;
            for (OWLClassExpression conjunct : ((OWLObjectIntersectionOf) concept).getOperands()) {
                int depth = getDepth(conjunct);
                if (depth > maximum) {
                    maximum = depth;
                }
            }
            return maximum;
        }
        if (concept instanceof OWLObjectSomeValuesFrom) {
            return getDepth(((OWLObjectSomeValuesFrom) concept).getFiller()) + 1;
        }
        return 0;
    }
    
    public static OntologyStatistics getStatistics(Set<OWLSubClassOfAxiom> axioms) {
        return new OntologyStatistics(axioms.size(), getSubconcepts(axioms).size(), getConceptNames(axioms).size());
    }
    
    public static boolean subsumes(OWLClassExpression lhs, OWLClassExpression rhs) {
        boolean subsumes = true;
        Iterator<OWLClassExpression> i = rhs.asConjunctSet().iterator();
        while (subsumes && i.hasNext()) {
            OWLClassExpression c = i.next();
            subsumes = subsumesRhsNotIntersection(lhs.asConjunctSet(), c);
        }
        return subsumes;
    }
    
    private static boolean subsumesRhsNotIntersection(Set<OWLClassExpression> lhsConcepts, OWLClassExpression rhs) {
        if (isConceptName(rhs)) {
            return rhs.isOWLThing() || lhsConcepts.contains(rhs);
        } else { // must be an existential
            OWLObjectSomeValuesFrom rhsEx = (OWLObjectSomeValuesFrom) rhs;
            for (OWLClassExpression c : lhsConcepts) {
                if (isExistentialRestriction(c)) {
                    OWLObjectSomeValuesFrom lhsEx = (OWLObjectSomeValuesFrom) c;
                    if (lhsEx.getProperty().equals(rhsEx.getProperty())) {
                        if (subsumes(lhsEx.getFiller(), rhsEx.getFiller())) {
                            return true;
                        }
                    }
                }
            }
            return false;
        }
    }
    
    public static Set<OWLClass> getCNsFromConcept(OWLClassExpression rhs) {
        Set<OWLClass> result = new HashSet<>();
        for (OWLClassExpression concept : ELUtils.getSubconcepts(rhs)) {
            if (concept instanceof OWLClass) {
                result.add((OWLClass) concept);
            }
        }
        return result;
    }
}
